""" STIS-specific functions

"""



import numpy as np

from ....main import util

from ....engine import eetable
from ....engine import instrument
from ....engine import initialize
from ....engine.collection import AbstractCollection
from ....engine.geometry import Square, Point
from ....engine.spectrograph import Spectrograph
from ....engine.observed import SpecObservedItem
from ....engine.extracted import ExtractedItem, ExtractedCollection
from ....engine.noise import ImagDetectedNoiseItem
from ....engine._modified_observation import _ModifiedObservation
from ....engine.string_constants import TARGET, BACKGROUND, STRAY_LIGHT, TARGET_SCATTERED_LIGHT
from ....engine.exceptions import PyetcException

from pysynphot.spectrum import ArraySourceSpectrum


def get_configuration_data(ei):
    """ Load instrument and detector data from config files.

        This overrides the base function in order to handle
        STIS specifics.

    Parameters
    ----------
    ei: dict
      engine input

    Returns
    -------
    iconfig: dict
      dictionary of instrument data
    dconfig: dict
      dictionary of detector data
    custom_config: dict
      dictionary with custom data

    """
    # get baseline configuration data.
    iconfig, dconfig, custom_config = initialize.get_configuration_data(ei)

    # in spectroscopy mode, pixel sizes depend on spectral element. The
    # spectral element name in the input dict is also STIS-specific.
    # Furthermore, only the hi-res spectroscopic dispersers (which are
    # supported only on the MAMAs) have entries
    # for spectral pixel height and width; the others default to the
    # plain old pixel_height/width entries in the dict.
    if instrument.find_instrument_type(ei['science_mode']) == instrument.spectrograph_type:

        if 'spectral_pixel_height' in dconfig:
            dconfig['pixel_height'] = dconfig['spectral_pixel_height'].get(iconfig['spec_element'],
                                                                           dconfig['pixel_height'])
            
            dconfig['pixel_width'] = dconfig['spectral_pixel_width'].get(iconfig['spec_element'],
                                                                          dconfig['pixel_width'])


    # TODO: Is this the right place to get obswave?
    #        # Echelle factors are indexed by grating; the scattering factor
    #        # must be interpolated at obswave.
    #        obswave = ei['obswave']

    else:
        #there is no crsplit / nread functionality available for a MAMA detector, #491
        if "mama" in ei['detector']:
            iconfig['nreads'] = 1

    # spectral resolution must be returned in the detector config dict.
    dconfig['spectral_resolution'] = iconfig[dconfig['detector_name']]['spectral_resolution']

    # Gain in target acquisition mode is set to pre-defined value; thus
    # we should ignore whatever is specified in the input dict.
    if 'acq' in iconfig['science_mode']:
        try:
            dconfig['gain'] = dconfig['gain_targetacq']
        except KeyError:
            pass

    # FUV glow region or CCD dark level selected by the user must be
    # passed to the instrument custom configuration code.
    if 'fuvglowregion' in ei:
        custom_config['fuvglowregion'] = ei['fuvglowregion']
    if 'ccddarklevel' in ei:
        custom_config['ccddarklevel'] = ei['ccddarklevel']

    return iconfig, dconfig, custom_config


def get_detector_tables(instrument_name, detector_name, science_mode, spec_element):
    """ Gets enclosed energy tables for STIS spectroscopy.

    These tables are indexed by a combination of detector name and grating name.
    This function handles this special case of table indexing by constructing
    this composite string.

    Parameters
    ----------
    instrument_name: string
      the instrument name
    detector_name: string
      the detector name. Not used.
    science_mode: string
      the science mode.
    spec_element: str
      the 'disperser' entry in the engine input dictionary. An example would
      be 'fuvmama_0g140m'.

    Returns
    -------
    result: dict
        Dictionary with the enclosed energy tables associated to this specific disperser.

    """
    if spec_element:
        # Create the composite string that it wants
        detector_disperser = '_0'.join([detector_name, spec_element])
        result =  eetable.get_detector_tables(instrument_name, detector_disperser, science_mode)
    else:
        result = eetable.get_detector_tables(instrument_name, detector_name, science_mode)

    return result


def modify_configuration(instrument, custom_config):
    """ The instrument instance is populated with
        custom configuration attributes.

        Here we replace the threshold value in
        the brightbinpix Limit object by a
        custom-calculated value.

    Parameters
    ----------
    instrument: Imager or Spectrograph
      Instance being configured
    custom_config: dictionary
      custom configuration info - not used for now

    """
    binning = instrument.binx * instrument.biny

    # we need the ugly code below since the static nature of the Limits
    # object does not allow dynamically calculated thresholds.
    #
    # TODO: add on-the-fly threshold computation capability to Limit class.
    if instrument.detector.name == 'ccd':
        gain = instrument.detector.element.gain

        if binning > 1:
            result = 34300 - 1300 * binning
            if gain == 4.0:
                result = 4.0 * (65536 - 1600 * binning)

            # force dynamically calculated threshold value into
            # statically defined limit object so the threshold
            # can be used later to generate warnings.
            (instrument.limits['brightbinpix']).threshold[int(gain)] = result
        else:
            # set limit to infinite so the binned pixel limit won't be triggered ever
            (instrument.limits['brightbinpix']).threshold[int(gain)] = np.inf


def _init_custom_config(instr, instrument_type, iconfig, dconfig, custom_config):
    """
    Add custom configuration data to the instrument.custom Struct Object

    This is used in STIS:

    - in all MAMA modes, to initialize the buffer time factor.
    - in spec mode, to retrieve the echelle scattering factors for the particular
      detector and grating being used.
    - in all FUV MAMA modes, to add the rate from the glow region to the dark rate.
    - in all CCD modes, to find the dark rate.

    """
    if instr.detector.name in ['fuvmama', 'nuvmama']:
        instr.custom.half_buffer = dconfig['half_buffer']

    if instrument_type == instrument.spectrograph_type:

        detector_data = iconfig[instr.detector.name]

        if 'echelle_scatter' in detector_data and 'echelle_global' in detector_data:

            element = instr.spectral_element

            if element in detector_data['echelle_scatter']:
                instr.custom.echelle_scatter = detector_data['echelle_scatter'][element]

            if element in detector_data['echelle_global']:
                instr.custom.echelle_global = detector_data['echelle_global'][element]

    if instr.detector.name == 'fuvmama':
        user_selected_glow_level = custom_config['fuvglowregion']
        glow_rate = dconfig['dark_glow_region'][user_selected_glow_level]
        instr.detector.element.dark_rate_per_pixel += glow_rate

    if instr.detector.name == 'ccd':
        user_selected_dark_level = custom_config['ccddarklevel']
        dark_rate = dconfig['dark_current_rate'][user_selected_dark_level]
        instr.detector.element.dark_rate_per_pixel = dark_rate

    return instr


def _interpolate_echelle_factor(scatter_dict, obswave):
    """ Simple 1-D interpolation of the echelle scattering factor.

    Parameters
    ----------
    scatter_dict: dict
       dictionary with wavelength,value pairs
    obswave: float
       the wavelength where to interpolate

    Returns
    -------
    the interpolated value (float)

    """
    try:
        data = []
        xref = np.sort(list(scatter_dict.keys()))

        for k in xref:
            data.append(scatter_dict[k])

        return util.linear_interp(xref, data, obswave)
    # Handle errors for the benefit of the developer.
    # There is nothing the pyetc user can do at this point.
    except Exception as e:
        raise PyetcException("Echelle scattering interpolation error - ", e)


def modify_noise(instrument, noise_collection):
    """ The inter-order echelle scattered light is accounted
        for by modifying the Poisson noise associated to
        the target.

    Parameters
    ----------
    instrument: Spectrograph
       contains the factor used to compute the inter-order
       scattered light component coming from the target.
    noise_collection: NoiseCollection
       the noise to be modified in place

    Returns
    -------
    None

    """
    if not isinstance(instrument, Spectrograph):
        return

    if not hasattr(instrument.custom, 'echelle_scatter'):
        return

    target_noise = noise_collection[TARGET]

    # get attributes from original extracted item; they
    # are needed in order to build a new "extracted"
    # item for the scattered light noise component.
    orig_extracted_item = target_noise._extracted_item

    rate = orig_extracted_item._rate
    eefrac_used = orig_extracted_item._eefrac_used
    qyc_function = orig_extracted_item._qyc_function
    wavelength = orig_extracted_item._wavelength
    extraction_shape = orig_extracted_item.extraction_shape
    name = orig_extracted_item.name

    # Interpolate to get scattering factor from dict.
    scattering_factor = _interpolate_echelle_factor(instrument.custom.echelle_scatter, wavelength)

    # The config file defines the scattering factor as a multiplicative factor. This
    # factor (the inverse of the actual values in the config file) is to be applied
    # to the target rate, in order to boost the number of counts from the target to a
    # level appropriate to describe the extra noise coming from scattered light. This
    # of course assumes an implicit Poisson distribution for this noise term. The actual
    # distribution is irrelevant, since the scattering model was *defined* in that way.
    #
    # Here, we instead subtract 1.0 from the original (inverted) value to create an *additive*
    # noise term to be added to the noise collection. This term also describes the extra
    # noise in the calculation coming from the scattered light, but is separate from the
    # actual target noise term. Again, it is assumed that this additional noise term has
    # a Poisson distribution. This format is convenient because it allows separate reporting
    # of each noise contribution.
    scattered_light_rate = (1. / scattering_factor - 1.) * rate

    # Finally, build new extracted item instance, new associated
    # noise item instance, and stick it into the noise collection.
    scattered_light_item = ExtractedItem(TARGET_SCATTERED_LIGHT,
                                         scattered_light_rate,
                                         extraction_shape,
                                         eefrac_used=eefrac_used,
                                         qyc_function=qyc_function,
                                         wavelength=wavelength)

    modified_target_noise = ImagDetectedNoiseItem(TARGET_SCATTERED_LIGHT, scattered_light_item)

    noise_collection.add(modified_target_noise)


def modify_total(instrument, total):
    """ The DetectedCollection with total signal+background that is
        used by the feasibility thread must be augmented with the stray
        light component associated with the global echelle scattering
        factor.

    Parameters
    ----------
    instrument: Spectrograph
       contains the factor used to compute the global scattered light
       component coming from the target.
    total: DetectedCollection
        on exit, updated with the global scattered stray light component.

    Returns
    -------
    None

    """
    # These conditions preclude computation of the
    # echelle global scattered light.

    if not isinstance(instrument, Spectrograph) or \
       not hasattr(instrument.custom, 'echelle_global') or \
       instrument.central_wavelength not in list(instrument.custom.echelle_global.keys()):
        return

    # Can compute echelle global scattered light, so, proceed...

    target = total[TARGET]
    background = total[BACKGROUND]

    scattering_factor = instrument.custom.echelle_global[instrument.central_wavelength]
    waves = target.binwave
    factors = np.ones(shape=waves.shape)
    # The -1 converts the multiplicative scattering factor as originally
    # defined by the instrument team, to a factor that can be used to
    # compute an additive stray light component.
    factors *= float(scattering_factor) - 1.

    # Build an Observation with the scattering factors.
    scatter_factor = ArraySourceSpectrum(wave=waves, flux=factors, name=STRAY_LIGHT)
    scatter_factor = _ModifiedObservation(scatter_factor, 1.0, 'multiply', binset=waves)

    # Scattered stray light comes from the target flux.
    # Build an Observation with the stray light component.
    stray_light_observation = _ModifiedObservation(target.observation, scatter_factor, 'multiply',
                                                   binset=target.binwave, force='taper')
    qyc_function = instrument.detector.get_qyc_function()

    # POINT TARGET:
    # The global scattering factor, as defined by the instrument
    # team, contains already an "hidden" area. I am guessing that
    # happens because one is using a point target to derive flux
    # that is extended over an area. Since this area, whatever it is,
    # was already accounted for in the scattering model that defined
    # the scattering coefficient, we have to prevent any code from
    # here downstream to multiply the flux of this additive stray
    # light component by any non-unit area. In other words, normally
    # a stray light component would consist of a flux per unit area,
    # multiplied by an area. In this particular case however, the flux
    # was already expressed in flux units, thus we associate to it an
    # unit area. Any shape would do; we use a square for simplicity.
    #
    # EXTENDED TARGET:
    # The target flux is expressed in flux per unit area units, thus we
    # associate to the stray light component the actual visible area
    # of the target.

    if isinstance(target.visible_shape, Point):
        stray_light_visible_shape = Square(1., 1., 1., 1., refers_to=STRAY_LIGHT)
    else:
        stray_light_visible_shape = target.visible_shape

    stray_light = SpecObservedItem(STRAY_LIGHT, instrument,
                                   stray_light_observation,
                                   stray_light_observation,
                                   stray_light_visible_shape,
                                   qyc_function,
                                   warnings=target._warnings)
    background.add(stray_light)


def _remove_stray_light(extracted_collection, instrument):

    """ The stray light component associated with the echelle
        global scattering can not contribute to the brightest
        pixel calculation. This function creates a copy of
        the input collection with the stray ligth component
        removed. This is supposed to be used only on
        collections that contain brightest pixel data.
    """
    if not hasattr(instrument.custom, 'echelle_global'):
        return extracted_collection

    # Copy each item in the input collection to the
    # output collection, as is. The exception is
    # the BACKGROUND item, which is a collection
    # to be processed separately.
    contents = []
    for item in extracted_collection:
        if item.name != BACKGROUND:

            contents.append(item)

        else:

            background = extracted_collection[BACKGROUND]

            # Build a BACKGROUND collection to be added to
            # the output. This collection has all the items
            # in the input BACKGROUND but the STRAY_LIGHT
            # item.

            background_contents = []
            for background_item in background:
                if background_item.name != STRAY_LIGHT:
                    background_contents.append(background_item)

            new_background_collection = ExtractedCollection(BACKGROUND + '_modified',
                                                            contents=background_contents)
            contents.append(new_background_collection)

    result = ExtractedCollection(extracted_collection.name + '_modified', contents=contents)

    return result


def _double_observation(extracted_spectrum):
    """  Convenience function that doubles an ExtractedSpectrum
         instance in place by adding its internal _ModifiedObservation
         instances to themselves. This is a workaround for the
         limitation that _ModifiedObservation implements __add__
         but not __mul__
    """
    extracted_spectrum.rebinned_observation = extracted_spectrum.rebinned_observation + \
                                              extracted_spectrum.rebinned_observation
    extracted_spectrum.reduced_observation = extracted_spectrum.reduced_observation + \
                                             extracted_spectrum.reduced_observation


def _dispersion_binning_correction(extracted_collection, binx):
    """ This is a rather arbitrary function that applies the binning
        correction in the dispersion direction (binX) by just
        multiplying all the elements in the extracted collection
        by the binning factor.

        This is the way the old (and Java) code work, and for now
        we aim to only emulate their behavior. It is reasonable to
        expect that such multiplication would be somehow paired with
        (or replaced by) a re-binning of the wavecat associated with
        the Observation instances, but that doesn't happen anywhere
        in the old code.

        The old code has a rather convoluted way of applying the
        binning corrections in spatial and dispersion directions.
        One side effect is that corrections end up applyed differently
        for slitted and slitless modes. Another bug relates to biny
        only: it's applied twice in some circumstances, and just once
        in others. This behavior is not currently emulated by this
        function.

        The binning correction in the dispersion direction could be
        less trivial than the current algorithm seems to imply. It
        could in principle be applied in different ways depending on
        what exactly is the source of the extraction. Target and sky
        could perhaps be handled in a different way than dark and
        thermal. Slits narrower than one binned pixel in the dispersion
        direction could also create problems for point targets,
        extended targets, and sky extractions, in possibly different
        ways in each case.

        Of course, this function should be called only by
        spectroscopic-specific code. It is also supposed to be used
        only on collections that contain brightest pixel data.
    """
    if binx == 1:
        return extracted_collection

    # Note that this works only for binx = 2 and binx = 4.
    # These two values are arbitrary (from the ETC viewpoint)
    # and enforced at the STIS spectroscopic input page level.
    # It's unlikely that the STIS CCD will ever be operated
    # with different settings, but in case that happens, this
    # code must be modified accordingly.

    for count in range(0, binx, 2):
        for item in extracted_collection:
            if isinstance(item, AbstractCollection):
                for internal_item in item:
                    _double_observation(internal_item)
            else:
                _double_observation(item)

    return extracted_collection


def modify_one_pixel(instrument, extracted_collection):
    """ Modifies an ExtractedCollection with data for
        brightest pixel calculations:

        - removes stray light component associated with
          the echelle global scattering factor.
        - applies correction for binning in the
          dispersion direction.

    Parameters
    ----------
    extracted_collection: ExtractedCollection
      the collection to be modified

    Returns
    -------
    result: ExtractedCollection
      a modified copy of the input collection, or
      a  reference to the input collection if no
      change took place.

    """
    if not isinstance(instrument, Spectrograph):
        return extracted_collection

    collection = _dispersion_binning_correction(extracted_collection, instrument.binx)

    return _remove_stray_light(collection, instrument)


def prepare_vectors_for_plotting(target, background):
    """ Modifies target and/or background vector objects according
    to the expectations of what the plots and table should do.

    For STIS, this looks for the presence of a stray light element
    in the background collection, and, if found, remove it. Plots
    and tables should not include the echelle scattered light
    contribution to the background.

    Parameters
    ----------
    target: ObservedItem
       the observed target
    background: Collection
       the observed background collection

    Returns
    -------
    None
    """
    if STRAY_LIGHT in background:
        background.remove(STRAY_LIGHT)


def compatibility(engine_input, result):
    """ Handles the slit specification in STIS spec mode.

    Parameters
    ----------
    engine_input: dict
      the engine input
    result: dict
      the resulting engine input after compatibility changes are applied

    Returns
    -------
    None
    """
    # Remove slit part from obsmode string, IFF it is a lossy slit.
    # This assumes that the slit specifier always has
    # an 'x' in it, and that nothing else does. Both seem true for STIS.
    tokens = engine_input['obsmode'].split(',')
    slits = [t for t in tokens if t.lower().find('x') > 0]
    newobm = engine_input['obsmode']
    if slits: # has a slit as a token
        slit = slits[-1]
        # only remove the lossy slits
        if "ND" not in slit.upper() and not slit.upper().startswith('F'):
            tokens.remove(slit)
            newobm = ','.join(tokens)
    result['obsmode'] = newobm


def compute_buffer_time(instrument, detector_total_rate):
    """
    Implements buffer time calculation for STIS

    """
    if hasattr(instrument.custom, 'half_buffer') :
        result = instrument.custom.half_buffer / detector_total_rate
        instrument.custom.output.buffer_time = result

    return instrument






# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A set of exceptions for pandeia to raise

The web interface will recognize any of these exceptions and produce a
pretty and user-friendly (to the extent that you gave good text) error
message.

"""
import re


class PandeiaException(Exception):

    """superclass of all custom pandeia exceptions

    This exception contains these fields of interest:

        message - the message provided by the code that raised the exception

        results - the results dict that the engine was working on at the
            time of the exception (this may be an empty dictionary if the
            engine code cannot provide it)

    """

    def __init__(self, value='Unspecified Error', results=None):
        Exception.__init__(self, value)
        self.results = results
        self.name = "Unspecified Pandeia exception"

    def addinfo(self, more_info):
        """ A convenience method so a calling function can add additional
        information to a raised exception before reraising.
        """
        args = list(self.args)
        args[0] = str(args[0]) + " " + str(more_info)
        self.args = tuple(args)

        # self.message is deprecated as of python 2.6, but we must set it
        # now in order to get the desired behavior.
        self.message = self.args[0]

# Errors resulting not from user inputs, but coding or connection problems.

class EngineOutputError(PandeiaException):

    """
    an engine error that is probably related to engine output

    raise engine.exceptions.OutputError( 'descriptive text', result_dict )

    The result dict is the partially completed result dictionary that the
    engine would have returned if it could finish the calculation.

    This error should include the etcid that issued it

    As used, these are problems with bad data in $pandeia_refdata and engine coding.

    """

    def __init__(self, value="Output Error", results=None):
        PandeiaException.__init__(self, value)
        self.name = "Output Error"

class InternalError(PandeiaException):

    """
    an error that should not have happened and is probably a bug
    We should include the etcid in the error information?
    """

    def __init__(self, value="Internal Error", results=None):
        PandeiaException.__init__(self, value)
        self.name = "Internal Error"

class UnsupportedError(PandeiaException):

    """ Certain cases are 'available' but 'unsupported'.
    Raise this exception for those cases.
    """

    def __init__(self, value, results=None):
        PandeiaException.__init__(self, value, results)
        self.name = "Unsupported Option Selected"

class DataConfigurationError(PandeiaException):

    """
    an engine error that is probably related to configuration data

    raise engine.exceptions.DataConfigurationError( 'descriptive text', result_dict )

    The result dict is the partially completed result dictionary that the
    engine would have returned if it could finish the calculation.

    This error should include the etcid that issued it

    """

    def __init__(self, value="Data Configuration Error", results=None):
        PandeiaException.__init__(self, value)
        self.name = "Data Configuration Error"


class DataError(PandeiaException):

    """'
    Some data entity which should exist cannot be found
    """

    def __init__(self, value="Data Error", results=None):
        PandeiaException.__init__(self, value)
        self.name = "Data Error"

# ### USER INPUT ERRORS ### #

class UserInputError(PandeiaException):
    """
    Parent class for all pandeia exceptions that arise from improper input data.
    These are errors that simply retrying a calculation will not fix.
    """

    def __init__(self, value="User Input Error", results=None):
        PandeiaException.__init__(self, value)
        self.name = value


class EngineInputError(UserInputError):

    """
    an engine error that is probably related to user input

    raise engine.exceptions.EngineInputError( 'descriptive text', result_dict )

    The result dict is the partially completed result dictionary that the
    engine would have returned if it could finish the calculation.

    This error should include the etcid that issued it

    """

    def __init__(self, value="Input Error", results=None):
        UserInputError.__init__(self, value)
        self.name = "Input Error"


class SynphotError(UserInputError):

    """ Any problem that arises from synphot should raise
    this exception
    """

    def __init__(self, value, results=None):
        # Inspired by https://stackoverflow.com/questions/50618116/regex-for-finding-file-paths
        # this will only work for file paths on UNIX systems - not Windows
        # It will also attempt to remove (and therefore mangle) URLs, however
        # at the moment there are no known Synphot errors containing URLs.
        fixvalue = re.sub(r"(\/.*/)", "", value)
        # Synphot errors are very long, but their first line contains all we generally need to know
        fixvalue = fixvalue.split("\n")[0]
        UserInputError.__init__(self, fixvalue, results)
        self.name = "Synphot Error"

class RangeError(UserInputError):

    """ Any problem caused by some value (usually a wavelength) falling
    outside a valid range should raise this exception.
    """

    def __init__(self, value, results=None):
        UserInputError.__init__(self, value, results)
        self.name = "Value Out of Range"


class WavesetMismatch(UserInputError):

    """ Raised when attempting to combine spectra of different
    wavelength sets
    """

    def __init__(self, value, results=None):
        UserInputError.__init__(self, value, results)
        self.name = "Inconsistent wavelength arrays"

# ### UNUSED PYETC EXCEPTIONS ### #

#class BadEnumeration(PandeiaException):
#
#    """
#    use this when an input field is expected to be one of a list of
#    known values, but it is not.
#    """
#
#    def __init__(self, value=None, results=None, field=None):
#        PandeiaException.__init__(self, value, results)
#        self.name = "Unrecognized Option"

# Make new exceptions here, if necessary.  Subclasses of InputError
# and InternalError require nothing special.  New subclasses of
# PandeiaException require new code in etc_web/etc/views.py

#class CalculationError(PandeiaException):
#
#    """ A catch-all exception for otherwise nonspecific exceptions
#    that are raised by the engine.
#    """
#
#    def __init__(self, value=None, results=None):
#        PandeiaException.__init__(self, value, results)
#        self.name = "Calculation Error"

# ### BACKGROUND ERRORS ### #

class BackgroundError(PandeiaException):

    """
    Raised when the Background Butler runs into errors.

    These errors could include:
        - invalid (position, date) combination
        - no stray light data available

    Usage:
        raise engine.custom_exceptions.BackgroundError('descriptive text')
    """

    def __init__(self, value='Background Error', debug=None):
        PandeiaException.__init__(self, value)
        self.name = value
        if debug:
            print('Background Error:')
            print(debug)


class StraylightPositionError(BackgroundError, UserInputError):

    """
    Raised when the specified position (ra, dec) is not observable on the
    specified date.

    Usage:
        raise engine.custom_exceptions.StraylightPositionError('descriptive text')
    """

    def __init__(self, value='User Input Related Background Error', debug=None):
        BackgroundError.__init__(self, value, debug)
        UserInputError.__init__(self, value)
        self.name = value


class StraylightDataError(BackgroundError):

    """
    Raised when needed stray light data is missing.

    Usage:
        raise engine.custom_exceptions.StraylightDataError('descriptive text')
    """

    def __init__(self, value='Background Error', debug=None):
        BackgroundError.__init__(self, value, debug)
        self.name = value


class DatelessBGError(BackgroundError):

    def __init__(self, value='Dateless Background Error', debug=None):
        BackgroundError.__init__(self, value, debug)
        self.name = value


class BMGError(BackgroundError):

    """
    Raised when the BMG returns an unhappy status code.

    Usage:
        raise engine.custom_exceptions.BMGError('status code')
    """

    def __init__(self, status, ra_dec_str, date_str):

        # these are the codes we can share with the user
        public_codes = {
            '500': 'BMG server error.',
            '501': 'Date (%s) out of range.' % date_str,
            '502': 'Ra/Dec (%s) out of range.' % ra_dec_str,
            '503': 'Wavelength out of range.'
        }

        # these are codes that we will log but won't share with the user
        private_codes = {
            '400': 'Empty bmg block (e.g. nothing after bmg keyword)',
            '401': 'No bmg block (e.g. no bmg keyword)',
            '402': 'Bad query block (e.g. badly formatted block)',
            '403': 'Bad query_type value (e.g. miss-spelled or un-supported)',
            '404': 'Empty query_type value (e.g. missing query_type value)',
            '405': 'No query_type keyword in query block (e.g. miss-spelled)',
            '504': 'Kernel list is inaccessible',
            '505': 'File wainscoat.txt is inaccessible',
            '506': 'Schlegel interstellar dust map fits file not found or inaccessible',
            '507': 'Schlegel interstellar dust map file name too long',
            '508': 'Error reading the std_spectrum_wavelengths.txt file',
            '509': 'Wavelength outside (0.1, 1200) micros for CIB',
            '510': 'Cib_spec_mode returns -1, wl outside (0.1, 1200)',
            '511': 'Error in opening file starcount.txt (when environment variable OUTPUT_STARCOUNT is set)'
        }

        if status in public_codes:
            # pass the error message to the user
            BackgroundError.__init__(self, public_codes[status])
        elif status in private_codes:
            # log the actual error, but just show the user something generic
            msg = 'The background model server encountered an unexpected error.'
            BackgroundError.__init__(self, msg, private_codes[status])
        else:
            # Unexpected/unhandled error, but pass it on
            BackgroundError.__init__(self, "BG error with unknown status of: "+str(status))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .config import DefaultConfig
from .custom_exceptions import EngineInputError
from .scene import Scene
from .strategy import StrategyFactory


class Observation(DefaultConfig):

    """
    This class collects the pieces required to make an ETC simulated observation.
    """

    def __init__(self, scene=None, instrument=None, strategy=None, config={}, webapp=False, **kwargs):
        DefaultConfig.__init__(self, config=config, webapp=webapp, **kwargs)

        # make sure we're given a properly configured Instrument
        if instrument is not None:
            self.instrument = instrument
            self.mode = self.instrument.mode
        else:
            msg = "Must provide configured Instrument subclass."
            raise EngineInputError(value=msg)

        # Initialize the observation. This either uses a preconfigured scene, or 
        # requires webapp==False to load a default scene and source.
        if scene is None and not webapp:
            self.scene = Scene(instrument.telescope.tel_name, webapp=webapp)
        else:
            self.scene = scene

        # use strategy we're given or fall back to default for the given instrument if webapp==False
        if strategy is not None:
            self.strategy = strategy
        elif not webapp:
            self.strategy = StrategyFactory(self.instrument, webapp=webapp)
        else:
            msg = "Must provide valid Strategy."
            raise EngineInputError(value=msg)

    def get_random_seed(self):
        """
        Return the configured seed for the random number generator.
        """
        return self.random_seed



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import bisect
import warnings

import numpy as np
import numpy.ma as ma
import scipy.interpolate as interp
from .custom_exceptions import EngineInputError, DataConfigurationError
from .constants import MIN_CLIP
from .exposure_factory import ExposureFactory
from .utils import spectrum_resample
from .debug_utils import timethis

class Detector:

    def __init__(self, config, webapp=False):
        """
        All functions that are detector-specific will live under children of this class

        DetectorNoise will use the this class (or rather, its detector-specific subclasses) to produce noise.
        """
        self.webapp = webapp
        self.detector_config = config["detector_config"]
        self.subarray_config = config["subarray_config"]
        self.readout_pattern_config = config["readout_pattern_config"]
        self.input_detector = config["input_detector"]

        # slopemode is given among the detector parameters that otherwise really define exposure
        self.slopemode = self.input_detector.get('slopemode', 'multiaccum')

        self.__dict__.update(self.detector_config)
        self.exposure_spec = self.get_exposure_pars(config)

    def get_exposure_pars(self, name="pattern_name"):
        """
        Define exposure parameters from instrument and environment.

        Parameters
        ----------
        name: str
            Optional name to assign ExposureSpecification instance

        Returns
        -------
        exposure_spec: pandeia.engine.exposure.ExposureSpecification instance
            Exposure parameters encapsulated within ExposureSpecification instance
        """
        # These are values that get set when the Instrument is instantiated, either from defaults
        # or passed configuration data.
        # These are defined in the reference data as "ramp_config" and apply to a whole instrument

        exp_conf = {"input_detector": self.input_detector, "subarray": self.subarray_config,
            "readout_pattern": self.readout_pattern_config, "detector_config": self.detector_config}

        exposure_spec = ExposureFactory(self.det_type, config=exp_conf, webapp=self.webapp)

        return exposure_spec

    def set_saturation(self, flag):
        # this function turns saturation on and off - if the flag is True, we 
        # compute saturation. If toggle is False, we do not.
        if not flag:
            self.fullwell = np.inf

    def set_readnoise(self, flag):
        # this function turns readnoise on and off - if the flag is False,
        # we set the readnoise to 0.
        if not flag:
            self.rn = 0

    def set_excess(self, flag):
        # this function turns the various excess noise components on and off - if the flag is False,
        # we set var_fudge, excessp1, and excessp2 to 0.
        if not flag:
            self.var_fudge = 1
            self.excessp1 = 0
            self.excessp2 = 0

    def get_slope_variance(self, rate, unsat):
        # This function allows different slope variance functions (used for MULTIACCUM TA)
        slope_var, slope_rn_var = self.slope_variance(rate, unsat)
        return slope_var, slope_rn_var

    def scale_exposure_time(self, extracted):
        raise NotImplementedError("Reverse ETC calculations for {} are not implemented.".format(type(self)))

    def set_time(self, new_time):
        raise NotImplementedError("Reverse ETC calculations for {} are not implemented.".format(type(self)))

    def to_resels_scalar(self, wave, values, refwave, squared=False):
        """
        Extract scalar value from array in resels.
        Requires an array, a resel (read from the detector), and the reference wavelength
        Uses astropy 1D box model to apply the resel value.
        The one advance this has over previous Pandeia behavior is that you DO get the (value) at the
        average position of your reference wavelength.

        This works in any case where you're dropping the wavelength dimension, so also cube->2D

        Parameters
        ----------
        wave: ndarray
            the wavelength array, in microns
        values: ndarray
            the value array 
        refwave: float
            the reference wavelength in microns
        squared: bool
            True in case the value array is a noise array that must be summed in quadrature (JETC-3020)

        Returns
        -------
        value: float
            the value in resels
        """
        values = ma.fix_invalid(values, fill_value=0.0).data
        if len(wave) > 1:
            if hasattr(self,"pixels_per_resel"):
                # find out what fractional index position the reference wavelength is at
                cenidx = np.interp(refwave, wave, np.arange(len(wave)))
                mask_pix = np.array([0.0, 0.0, 1.0, 1.0, 0.0, 0.0])
                mask = np.array(
                    [
                        0,
                        cenidx - (self.pixels_per_resel + 1) / 2.0,
                        cenidx - (self.pixels_per_resel - 1) / 2.0,
                        cenidx + (self.pixels_per_resel - 1) / 2.0,
                        cenidx + (self.pixels_per_resel + 1) / 2.0,
                        len(wave)-1
                    ]
                )
                boxfilter = np.interp(np.arange(len(wave)), mask, mask_pix)
                if len(values.shape) == 3: # if we're doing a cube, the boxfilter needs to be reshaped
                    boxfilter = boxfilter.reshape(-1,1,1)

                # sum the array, linearly, or in quadrature. This is used for noise, and
                # is effectively the same as multiplying the values by the square root of
                # the boxfilter, just as the noise extraction is over the square of the noise aperture.
                if squared:
                    resel_value = np.sum(values ** 2 * boxfilter) / \
                                  (self.pixels_per_resel ** 2)  # the 1D values are simply extracted over the pixel-area of one resel
                    resel_value = np.sqrt(resel_value)
                else:
                    resel_value = np.sum(
                        values * boxfilter) / self.pixels_per_resel  # the 1D values are simply extracted over the pixel-area of one resel

            else:
                # there are no resels but it is a 1D array
                wave_index = (np.abs(wave - refwave)).argmin()
                resel_value = values[wave_index]
        else:
            # it is a single-element array
            resel_value = values[0]

        return resel_value

    # TODO: Resels for downsampling a cube

    def to_resels_array(self, wave, values):
        """
        Convert array to resels.
        Requires an array and a resel (read from the instrument)

        Parameters
        ----------
        wave: ndarray
            the wavelength array, in microns
        values: ndarray
            the value array 

        Returns
        -------
        rebinwave: ndarray
            the rebinned wavelength array, in microns
        array: ndarray
            the rebinned array
        """
        if hasattr(self,"pixels_per_resel"):
            if self.pixels_per_resel != 1 and len(wave) > 1:
                # spectrum_resample cannot handle NaNs, so we need to save the nanmask and set nans to 0.
                nanmask = np.isfinite(values) 
                values = ma.fix_invalid(values, fill_value=0.0).data

                rebinwave = wave[::self.pixels_per_resel] # this is the same operation conducted by PyETC
                resel_array = spectrum_resample(values, wave, rebinwave)

                # rebin the nanmask (all elements in resel must be true= not NaN for the resel to not be NaN) and reapply
                rebinmask = [np.all(nanmask[x:x+self.pixels_per_resel]) for x in range(0, len(values), self.pixels_per_resel)]
                return rebinwave, resel_array*rebinmask
        # if none of the above were true, return the originals
        return wave, values

class Detector_MultiAccum(Detector):

    def get_unsaturated(self, slope):
        """
        Calculate the number of unsaturated groups in each pixel, given a specified slope
        and full well value. There is a minimum sensible value of unsaturated number of
        groups that defines full saturation. The default is 2, but this can potentially be
        set to a higher value, or even 1(!) for certain observing modes. It is not
        expected that a value of 0 can ever be anything but saturated.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        Returns
        -------
        unsat_ngroups: MaskedArray
            The number of unsaturated groups present in the ramp. The mask separates
            pixels that have full saturation from those that do not. The former will not
            have a valid noise measurement and cannot be used in a strategy.

        """

        # get the number of groups to saturation
        ngroups_to_sat = self.get_before_sat(slope)

        # the number of unsaturated groups is clipped by the ngroups_to_sat value for
        # every pixel
        max_ngroups = ngroups_to_sat.clip(0,self.exposure_spec.ngroup)

        # Mask values that are fully saturated - if the maximum before saturation is less
        # than self.mingroups, it's clearly saturated.
        unsat_ngroups = ma.masked_less(max_ngroups, self.mingroups)

        return unsat_ngroups

    def get_saturation_fraction(self, slope):
        """
        This method returns a map of the fraction of saturation that each pixel reaches, given the current
        exposure setup.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        """
        # we clip to 1e-10 to avoid division by zero errors even if the slope is tiny.
        return self.exposure_spec.saturation_time / (self.fullwell / slope.clip(MIN_CLIP, np.max([MIN_CLIP, np.max(slope)])))

    def _get_variance_fudge(self, wave):
        """
        Some instruments (e.g. MIRI) have measured noise properties that are not well explained
        or modeled by Pandeia's current noise model. This method returns the multiplicative fudge factor
        that is to be applied to the pixel rate variance to match IDT expectations. The fudge factor can be
        chromatic so needs to accept an array of wavelengths to interpolate over. By default this returns
        a 1's array the same length as wave scaled by the 'var_fudge' scalar value.

        Parameters
        ----------
        wave: 1D numpy.ndarray
            Wavelengths to interpolate chromatic variance over
        """
        if hasattr(self, 'var_fudge'):
            var_fudge = self.var_fudge
        else:
            var_fudge = 1
        
        return var_fudge * np.ones(len(wave))

    def get_before_sat(self, slope):
        """
        This method returns a map of the maximum number of groups each pixel can be exposed to, before saturation.
        The formula is calculated by equating the time-to-saturation to the saturation time formula and isolating
        ngroup. The resulting fractional value is then rounded down to an integer.

        These equations are derived by setting saturation_time = time_to_saturation in equation 7 (for NIR H2RG
        detectors) or equation 6 (for MIR SiAs detectors, now out of the timing document revision, JWST-STScI-006013-A)
        of the timing document, Valenti et al. 2017, JWST-STScI-006013, and solving for n_groups.

        Technically speaking, we are getting the groups BEFORE saturation by rounding the number of groups down to the
        nearest integer. If groups_before_sat is EXACTLY an integer (which is unlikely), we will get one partially
        saturated pixel.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        fullwell: positive integer
            The number of electrons defining a full well (beyond which the pixel reads become unusable for science
            due to saturation).
        """
        # we clip to 1e-10 to avoid division by zero errors even if the slope is tiny.
        time_to_saturation = self.fullwell / slope.clip(MIN_CLIP, np.max([MIN_CLIP, np.max(slope)]))

        groups_before_sat = self._get_groups_before_sat(time_to_saturation)

        # The groups_before_sat value should never be less than 0 - that would imply saturation before we started
        # collecting any photons.
        groups_before_sat = np.maximum(groups_before_sat,np.zeros_like(groups_before_sat))

        return np.floor(groups_before_sat)

    def slope_variance(self, rate, unsat_ngroups):
        """
        Calculate the variance of a specified MULTIACCUM slope.

        Inputs
        ------
        rate: ndarray
            The measured slope of the ramp (ie, the rate) per pixel
            This can be a one-dimensional wavelength spectrum (for a 1D ETC)
            or a three-dimensional cube ([wave,x,y] for a 3D ETC).

        unsat_ngroups: ndarray
            Number of unsaturated groups for each pixel

        Returns
        -------
        slope_var: ndarray
            Variance associated with the input slope.
        slope_rn_var: ndarray
            The associated variance of the readnoise only
        """
        # Rename variables for ease of comparison with Robberto 2009 eq. 35 (JWST-STScI-001853)
        # The noise calculation depends on the pixel rate BEFORE IPC convolution. fp_pix_variance takes that pre-IPC
        # rate and scales it by the quantum yield and Fano factor to get the per-pixel variance in the electron rate.
        variance_per_pix = rate['fp_pix_variance']

        # we discard any saturated groups. Copy unsat_ngroups, because we may modify it for rejected groups before using
        # it for noise calculations.
        n = ma.copy(unsat_ngroups)

        # removing with any pre- and post-rejected groups here. This is not stictly directly related to saturation, bu
        # behaves in a similar way (if there is <2 groups available, we cannot define a slope). The very bright regime
        # in which groups are not rejected can be implemented (in parts) by setting nprerej=npostrej=0.
        # Therefore we disable it here in that regime.
        if ((self.exposure_spec.nprerej != 0) or (self.exposure_spec.npostrej != 0)):
            # If a group is already rejected because of saturation, there is no need to reject another at the end of
            # the ramp.
            unsat_pixels = (self.exposure_spec.ngroup-np.ceil(n))<self.exposure_spec.npostrej
            n[unsat_pixels] -= self.exposure_spec.nprerej
            # we do have to reject any post-rejected frames from all pixels
            n -= self.exposure_spec.npostrej

        # H2RG/H4RG and SiAs detectors need different sample values. (Issue #2996)
        m, tframe = self._get_sample()

        tgroup = self.exposure_spec.tgroup

        # Compute the variance of a MULTIACCUM slope using Robberto's formula (eq. 35, JWST-STScI-001853). 
        # The rn_variance is also calculated using the unsaturated number of groups.
        slope_rn_var = self.rn_variance(unsat_ngroups=n)
        # The final slope variance slope may also be worse than the theoretical best value.
        # (also from Glasse et al. 2015, PASP 127 686).
        # the variance product contains the dark current, so we do not need to add it here.
        slope_var = (6. / 5.) * (n ** 2. + 1.) / (n * (n ** 2. - 1.)) * \
            (variance_per_pix / self.exposure_spec.tgroup) * \
            (1. - (5. / 3.) * (m ** 2. - 1.) / (m * (n ** 2. + 1.)) * 
            (tframe / self.exposure_spec.tgroup)) + slope_rn_var

        # The default fill value for masked arrays is a finite number, so convert to ndarrays, and fill with NaNs
        # to make sure missing values are interpreted as truly undefined downstream.

        slope_var = ma.filled(slope_var, fill_value=np.nan)
        slope_rn_var = ma.filled(slope_rn_var, fill_value=np.nan)

        return np.asarray(slope_var, dtype=np.float32), np.asarray(slope_rn_var, dtype=np.float32)

    def slope_variance_ta(self, rate, unsat_ngroups):
        """
        Calculate the slope variance for TA modes

        Inputs
        ------
        rate: ndarray
            Rate

        unsat_ngroups: ndarray
           The number of unsaturated groups.

           If unsat_ngroups not supplied, the approximation is that the read noise is
           negligible for pixels with signal rates high enough to saturate in part of the ramp.
           This is probably always a good approximation.

        Returns
        -------
        var_rn: ndarray if unsat_ngroups is supplied, float if unsat_ngroups is not supplied.
           Variance associated with the read noise only.

        """

        # Rename variables for ease of comparison with Robberto 2009 eq. 35 (JWST-STScI-001853)
        # The noise calculation depends on the pixel rate BEFORE IPC convolution. fp_pix_variance takes that pre-IPC
        # rate and scales it by the quantum yield and Fano factor to get the per-pixel variance in the electron rate.
        variance_per_pix = rate['fp_pix_variance']

        rn = self.rn
        n = unsat_ngroups  # we discard any saturated groups

        # H2RG/H4RG and SiAs detectors need different sample values. (Issue #2996)
        m, tframe = self._get_sample()

        nextract = self.exposure_spec.ngroup_extract

        # Throw error if the number of extract groups is >= number of groups.
        if self.exposure_spec.ngroup_extract > self.exposure_spec.ngroup:
            raise EngineInputError("Number of groups to extract, {}, must be less than or equal to the number of groups {}".format(
                self.exposure_spec.ngroup_extract, self.exposure_spec.ngroup))

        tgroup = self.exposure_spec.tgroup

        # The effective group time. This is the time from the first group to the next extracted group. For instance,
        # if ngroups=11 and nextract=3, textract = tgroup * 5
        textract = tgroup * (n-1)/(nextract-1)

        # The TA script will take the minimum of co-subtracted pairs. This has non-Gaussian error properties,
        # but a Monte Carlo Simulation indicates an improvement of 2/3 on the variance for the minimum of two
        # correlated pairs in a single ramp (as compared to a factor 1/2 for the average). It would be good to
        # demonstrate this analytically, but that has not yet been accomplished.

        # Number of possible pairs.
        npairs = nextract-1
        if npairs == 1:
            min_factor = 1.0
        elif npairs == 2:
            min_factor = 2./3.
        else:
            raise EngineInputError("Target acquisition currently only supports 2 or 3 extracted groups.")

        # Modification of MULTIACCUM formula for n=2 and a revised "effective" group time that takes into
        # account that only some groups obtained may be used to create the TA images.
        slope_rn_var = min_factor * (2*rn**2)/m/textract**2
        slope_var = min_factor * (variance_per_pix/textract) * (1.-(1./3.)*(m**2-1)/m * tframe/textract) + slope_rn_var

        # The default fill value for masked arrays is a finite number, so convert to ndarrays, and fill with NaNs
        # to make sure missing values are interpreted as truly undefined downstream.

        slope_var = ma.filled(slope_var, fill_value=np.nan)
        slope_rn_var = ma.filled(slope_rn_var, fill_value=np.nan)

        return np.asarray(slope_var, dtype=np.float32), np.asarray(slope_rn_var, dtype=np.float32)

    def rn_variance(self, unsat_ngroups=None):
        """
        Calculate the variance due to read noise only.

        Inputs
        ------
        rn: float
          Readnoise per pixel.

        unsat_ngroups: ndarray
           The number of unsaturated groups.

           If unsat_ngroups not supplied, the approximation is that the read noise is
           negligible for pixels with signal rates high enough to saturate in part of the ramp.
           This is probably always a good approximation.

        Returns
        -------
        var_rn: ndarray if unsat_ngroups is supplied, float if unsat_ngroups is not supplied.
           Variance associated with the read noise only.

        """
        if unsat_ngroups is None:
            n = self.exposure_spec.ngroup - (self.exposure_spec.nprerej + self.exposure_spec.npostrej)
        else:
            n = unsat_ngroups

        rn = self.rn

        # H2RG/H4RG and SiAs detectors need different sample values. (Issue #2996)
        m, tframe = self._get_sample()

        tgroup = self.exposure_spec.tgroup

        var_rn = 12. * rn ** 2. / (m * n * (n ** 2. - 1.) * tgroup ** 2.)

        # The readnoise on the slope may be worse than the theoretical best value
        # (see Glasse et al. 2015, PASP 127 686).
        if self.rn_fudge != 1:
            var_rn *= self.rn_fudge

        # Include the empirical correction for excess variance for long ramps
        # (Issue #2091)
        if self.excessp1 != 0.0 or self.excessp2 != 0.0:
            excess_variance = (12.0 * (n-1)/(n+1) * self.excessp1**2 - self.excessp2/np.sqrt(m)) / ((1 - n) * tgroup)**2
            var_rn += excess_variance
        return var_rn

    def _get_saturation_mask(self, rate):
        """
        Compute a numpy array indicating pixels with full saturation (2), partial saturation (1) and no saturation (0).

        Parameters
        ----------
        rate: None or 2D np.ndarray
            Detector plane rate image used to build saturation map from

        Returns
        -------
        mask: 2D np.ndarray
            Saturation mask image
        """
        saturation_mask = np.zeros(rate.shape)
        unsat_ngroups = self.get_unsaturated(rate)

        saturation_mask[(unsat_ngroups < self.exposure_spec.ngroup)] = 1
        saturation_mask[(unsat_ngroups < self.mingroups)] = 2

        return saturation_mask

    def calc_cr_loss(self, ngroups, pix_cr_rate):
        """
        Calculate the effective loss of exposure time due to cosmic rays. This uses the cosmic ray (CR) event rate
        that is contained within the telescope configuration to calculate the odds of a cosmic rays
        hitting a pixel. Pixels that are hit by cosmic rays are assumed to have ramps that are valid before the
        CR, but not after. The exposure specification is used to adjust the expectation value of the exposure
        time (i.e. ngroups) for a CR-truncated ramp.

        See discussion and links in:

        - https://confluence.stsci.edu/pages/viewpage.action?spaceKey=JWST&title=2014-11-10+Effect+of+CRs+on+SNR
        - Robberto (2010) Technical Report JWST-STScI-001928

        for more information on implementation and the numbers used.

        Parameters
        ----------
        ngroups: int
            Number of groups over which to calculate CR losses.  This will usually be number of unsaturated groups.

        Returns
        -------
        cr_ngroups: float
            This is the input ngroups scaled by the mean loss of time due to cosmic ray events.
        """

        # this is the average fraction of the ramp that is lost upon a CR event
        if ngroups < self.mingroups:
            # if for some reason (e.g. using single read noise model) ngroup is less than the minimum needed
            # to get a good ramp fit, then the whole ramp is lost upon a CR event
            ramp_frac = 1.0
        else:
            # the more groups you have per ramp, the less of the ramp you lose to CRs on average.
            # this formalism takes into account that there's always a fraction that's totally lost.
            # in the limit of infinite reads this converges on half the ramp since CRs are evenly
            # distributed.
            ramp_frac = 1.0 - 0.5 * (ngroups - self.mingroups) / (ngroups - 1.0)

        # the effective ramp exposure/measurement time, t_eff, is t_tot * (1 - ramp_frac * pix_cr_rate * t_ramp). We use the saturation
        # time as an approximation of the time during which a ramp can get damaged by a CR. This was previously more complex to
        # handle the case where a ramp is both saturated and hit by a CR, but that required a recalculation of the exposure time for
        # the unsaturated groups. We now simplify this (the difference is minimal and would add some complex logic now that the time
        # formulae are detector type dependent). This is simple.
        cr_ngroups = (1.0 - ramp_frac * pix_cr_rate * self.exposure_spec.saturation_time) * ngroups

        return cr_ngroups

    def get_slope_variance(self, rate, unsat):
        # We allow alternate ways to derive a signal from a detector slope,
        # but multiaccum is the default if no mode is provided. Slopemode
        # must be matched with a corresponding slope_variance function.
        if self.slopemode == "target_acq":
            slope_var, slope_rn_var = self.slope_variance_ta(rate, unsat)
        elif self.slopemode == "multiaccum":
            slope_var, slope_rn_var = self.slope_variance(rate, unsat)
        else:
            msg = "{} is not a valid slope mode.".format(self.slopemode)
            raise DataConfigurationError(value=msg)
            
        return slope_var, slope_rn_var

    def _get_groups_before_sat(self, time_to_saturation):
        return ((time_to_saturation / self.exposure_spec.tframe) - self.exposure_spec.nframe - self.exposure_spec.ndrop1) / \
                (self.exposure_spec.nframe + self.exposure_spec.ndrop2) + 1 

class Detector_UnevenMultiAccum(Detector_MultiAccum):

    def _power_value(self, snr):
        """
        Determine the optimal exponent P given a signal slope based on
        Table 2 of Casertano (2022) Roman-STScI-000394

        Code by S. Gomez Roman-STScI-000557
        Adapted to 2D arrays by A. Riedel
        """

        # Get highest SNR per pixel
        snr = np.max(snr,axis=0)

        power_grid=np.zeros_like(snr)
        power_grid[np.where((5 <= snr) & (snr <=  10))] = 0.4
        power_grid[np.where((10 <= snr) & (snr <=  20))] = 1.0
        power_grid[np.where((20 <= snr) & (snr <=  50))] = 3.0
        power_grid[np.where((50 <= snr) & (snr <= 100))] = 6.0
        power_grid[np.where(snr >= 100)] = 10.0

        return power_grid

    def slope_variance(self, rate, unsat_nframes, weights="optimal"):
        """
        The slope variance for an unevenly-sampled Multiaccum detector. Equations from
        Casertano 2022 (Roman-STScI-000394) via Gomez 2023 (Roman-STScI-000557)
        It is very similar to romanisim/ramp.py fit_ramps_casertano_no_dq() in RomanISim

        The basic difference is that the resultants, unlike JWST's groups, do not contain
        a constant number of frames. Therefore, we need to use a weighting scheme to
        accurately reproduce the noise.

        The complicated part of this is that each pixel may have a different number of
        valid resultants, and thus both nreads and ntotal are 3D arrays
        (spatial+resultants) where the third dimension is of different lengths per pixel.

        This can be handled by the unsat_resultants array containing just the number of
        unsaturated resultants - the computation is set up to stop there, per pixel.

        Worst case scenario, we can consider this function on a per-pixel basis, and run
        it X * Y times, vectorized.

        Inputs
        ------
        rate: (dict)
            Detected rate product
        unsat_nframes: (np.ndarray)
            2D array of the number of unsaturated frames per pixel
        weights: (str)
            Control for the weighting algorithm. One of "optimal", "simple", or "unweighted"
        """
        # first off, the inputs are actually frames; we need the resultants
        unsat_nresultants = self._frames_to_resultants(unsat_nframes)
        unsat_nresultants = np.asarray(unsat_nresultants, dtype=int)

        nreads = np.asarray(np.tile(np.reshape(self.exposure_spec.nreads, (-1, 1, 1)), unsat_nresultants.shape), dtype=float)
        ntotal = np.asarray(np.tile(np.reshape(self.exposure_spec.ntotal, (-1, 1, 1)), unsat_nresultants.shape), dtype=float)

        # This portion of the code is not from Casertano 2022 or RomanISIM (which handles cosmic rays by explicitly fitting partial ramps 
        # with custom Cython code)
        # Handle cosmic rays - they should not wipe out entire resultants, just make each one a little shorter.
        excess = np.vectorize(lambda a, b: self.exposure_spec.lastframe[a] - b)(unsat_nresultants, unsat_nframes)/self.exposure_spec.nresultants
        if excess.any() > 0:
            unsat_nresultants += 1
        # Remove the excess frames per pixel from the individual nreads and ntot entries for the last unsaturated resultant
        nreads -= excess
        ntotal -= excess

        # The source variance
        variance_per_pix = rate['fp_pix_variance']
        # add a dimension. Within this function, the dimensions are (resultant, y, x)
        variance_per_pix = np.expand_dims(variance_per_pix,0)

        tframe = self.exposure_spec.tframe

        # Calculate beginning time of each read
        t0 = np.insert(np.cumsum(ntotal, axis=0), 0, np.zeros_like(unsat_nresultants), axis=0)[:-1]
        # Calculate mean times for each resultant (Eq. 9)
        tmean = tframe * t0 + tframe * (nreads - 1) / 2 # would become negative if nreads is 0
        # convert electrons/s into electrons, per resultant
        var_per_resultant_per_pix = tmean * variance_per_pix


        if weights == "optimal":
            # Calculate weighting power from maximum snr (not flux)
            simple_snr = var_per_resultant_per_pix / np.sqrt(self.rn**2 + var_per_resultant_per_pix)
            weighting_power = self._power_value(simple_snr)
            # Add necessary dimension
            weighting_power = np.expand_dims(weighting_power,0)
            # Calculate optimal weights using Equation 45 in Casertano et al.
            # Midpoint of all resultants
            tmid = (tmean[0,:,:] + tmean[-1,:,:]) / 2
            w_ij = ((1 + weighting_power) * nreads) / (1 + weighting_power * nreads) * np.abs(tmean - tmid) ** weighting_power
        elif weights == "simple":
            # Calculate weights based on number of reads, or P = 0
            w_ij = nreads * np.ones_like(var_per_resultant_per_pix)
        elif weights == "unweighted":
            # Set all weights equal to 1
            w_ij = np.ones_like(var_per_resultant_per_pix)
        else:
            raise EngineInputError(f"Unrecognized weighting scheme {weights}.")

        # Calculate variance-based resultant times tau (eq. 15)
        tau = tmean - (nreads - 1) * (nreads + 1) * tframe / (6 * nreads)

        # compute the k_i weight matrix on a per-pixel basis to handle partial saturation
        # This algorithm is not from Casertano 2022 or RomanISIM though the equations are.
        k_i = np.zeros_like(w_ij)
        for y,val in enumerate(unsat_nresultants):
            for x,val in enumerate(unsat_nresultants[y]):
                values = self.weightcalc(unsat_nresultants[y,x],w_ij[:,y,x], tmean[:,y,x])
                k_i[:,y,x] = np.pad(values, (0,len(tmean)-len(values)), 'constant', constant_values=(0))

        slope_rn_var = self.rn_variance(k_i, ntotal)

        slope_var = np.zeros_like(rate['fp_pix_variance'])

        # Variance associated with the shot-noise.
        # Casertano+22 Eq 40
        V_s1 = np.sum(k_i ** 2 * tau, axis = 0)
        # sum needs to be 1*2 + 1*3 + 1*4 + 1*5, 2*3 + 2*4 + 2*5, 3*4 + 3*5, 4*5 and so on
        V_s2 = np.sum([np.sum([( k_i[i] * k_i[j] * tmean[i]) for j in range(i + 1, len(tau))], axis=0) for i in range(0, len(tau) - 1)], axis=0)
        slope_var  = V_s1 + 2 * V_s2

        # Casertano+22 Eq 41
        slope_var = slope_rn_var + slope_var * rate["fp_pix_variance"]

        mask = self._get_saturation_mask(rate['fp_pix_variance'])
        slope_var[mask==2] = np.nan 
        slope_rn_var[mask==2] = np.nan

        return slope_var, slope_rn_var

    def weightcalc(self, uns, w_ij, tmean ):

        # Calculate the auxiliary quantities W*t^0, W*t^1, W*t^2, and denomenator.
        # Casertano+22 eq 35
        # except in these arrays, the extra dimension is for unsat_resultants
        # these are resultants+1 because cosmic rays make this short by 1.
        f_0 = np.sum(w_ij[:uns], axis=0)
        f_1 = np.sum(w_ij[:uns] * tmean[:uns], axis=0)
        f_2 = np.sum(w_ij[:uns] * tmean[:uns] ** 2, axis=0)
        # Casertano+22 eq 36
        d   = f_2 * f_0 - f_1 ** 2

        # Casertano+22 eq 37
        k_i = (f_0 * tmean[:uns] - f_1) * w_ij[:uns] / d
        
        return k_i

    def get_unsaturated(self, slope):
        """
        Calculate the number of unsaturated resultants for each pixel, given a specified slope and full well value.
        There is a minimum sensible value of unsaturated number of resultants that defines full saturation.
        The default is 2, but this can potentially be set to a higher value, or even 1(!) for certain observing modes.
        It is not expected that a value of 0 can ever be anything but saturated.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        Returns
        -------
        unsat_nresultants: MaskedArray
            The number of unsaturated resultants present in the ramp. The mask separates pixels that have full saturation
            from those that do not. The former will not have a valid noise measurement and cannot be used in a strategy.

        """
        # get the number of groups to saturation
        nframes_to_sat = self._frames_before_sat(slope)

        # the number of unsaturated frames is clipped by the nresultants_to_sat value for every pixel
        unsat_nframes = nframes_to_sat.clip(0,self.exposure_spec.max_samples)

        return unsat_nframes
    
    def _frames_to_resultants(self, unsat_frames):
        """
        Utility class, translates unsaturated frames to unsaturated resultants

        We use the python bisect command to find where the number of frames would sort
        into the list of residual end-of-frames. If it sorts before the last frame of
        resultant X, it saturated during resultant X and we can only use X-1 resultants.
        As the array indicies are 0-based but resultants are 1-based, we don't need to
        subtract 1 and can simply use the value X.

        This code is not from Casertano 2022 or RomanISIM.

        Inputs
        ----------
        unsat_frames: np.ndarray
            2D array of the number of unsaturated frames per pixel

        Returns
        -------
        unsat_resultants: np.ndarray
            2D array of the number of unsaturated resultants per pixel
        """
        unsat = map(lambda i:bisect.bisect(self.exposure_spec.lastframe, i), unsat_frames.ravel())

        # reshape array
        # do not just turn the unsat (generator) into a list; it's a slow and expensive intermediate step
        unsat_nresultants = np.fromiter(unsat, np.dtype(float), count=unsat_frames.size).reshape(unsat_frames.shape)

        return unsat_nresultants

    def _get_saturation_mask(self, rate):
        """
        Compute a numpy array indicating pixels with full saturation (2), partial saturation (1) and no saturation (0).

        Parameters
        ----------
        rate: None or 2D np.ndarray
            Detector plane rate image used to build saturation map from

        Returns
        -------
        mask: 2D np.ndarray
            Saturation mask image
        """
        saturation_mask = np.zeros(rate.shape)
        unsat_nresultants = self.get_before_sat(rate)

        saturation_mask[(unsat_nresultants < self.exposure_spec.nresultants)] = 1
        saturation_mask[(unsat_nresultants < self.mingroups)] = 2

        return saturation_mask

    def rn_variance(self, k_i, total_frames=None):
        """
        Readnoise Variance

        Inputs
        ------
        rn: float
            Readnoise per pixel

        K_i: ndarray
            Weightmap for the readnoise contribution, per pixel

        total_resultants: ndarray
            Sparse 3D (along axis 0) array of frames per resultant

        Returns
        -------
        var_rn: ndarray
          Array of the unsaturated readnoise, which will vary based on the number of resultants.

        Args:
            K_i (_type_): _description_
            unsat_resultants (_type_, optional): _description_. Defaults to None.
        """

        # Variance associated with the read noise only
        # Casertano+22 eq 39

        # total_frames is really 1D, but it doesn't matter as the broadcasting shape is
        # valid and k_i implicitly handles all resultants that were saturated
        var_rn = np.sum(k_i ** 2 / total_frames, axis = 0) * self.rn**2

        return var_rn

    def get_slope_variance(self, rate, unsat):
        slope_var, slope_rn_var = self.slope_variance(rate, unsat)

        return slope_var, slope_rn_var
    
    def _frames_before_sat(self, slope):
        """
        To do the actual math of how many resultants we can have, we actually need the
        frames before saturation.

        The way this works is to compute the time to saturation using the slope and the
        fullwell depth. That time is then translated into numbers of frames (with the
        reference frame optionally included.)

        Inputs
        ------
        time_to_saturation: np.ndarray
            An array of the time (in seconds) to saturation, per pixel.

        tframe: float
            The time to read a single frame in the current setup.

        readout_pattern: list
            The actual unevenly-sampled readout pattern


        Returns
        -------
        unsat_resultants: np.ndarray
            An array showing how many resultants are unsaturated.

        """
        time_to_saturation = (self.fullwell / slope.clip(MIN_CLIP, np.max([MIN_CLIP, np.max(slope)])))

        # number of frames is total time / frame time. If reference frames are used, nreference will equal 1 and all of the
        # frame counts will be lower by one frame (the reference frame). A frame_before_saturation of -1 is possible if the
        # source is bright enough to saturate during the reference frame.
        frames_before_saturation = np.floor(time_to_saturation / self.exposure_spec.tframe) - self.exposure_spec.nreference

        return frames_before_saturation

    def get_before_sat(self, slope):
        """
        Because we need the intermediate products (frames) elsewhere, this API simply runs
        the two halves of the process.


        Inputs
        ------
        slope: (np.ndarray):
            2D array of electron variance rate, per pixel.

        Returns
        -------
        resultants_before_saturation: (np.ndarray)
            2D array of unsaturated resultants, per pixel.
        """

        frames_before_saturation = self._frames_before_sat(slope)

        resultants_before_saturation = self._frames_to_resultants(frames_before_saturation)

        return resultants_before_saturation

    def calc_cr_loss(self, nframes, pix_cr_rate):
        """
        Compute statistical cosmic ray losses, on the assumption that a cosmic ray hit
        (which on average will happen halfway through the ramp) will take out the entire
        remainder of the MA table.

        This is a direct adaptation of the MultiAccum cosmic ray equations to operate on
        individual frames.

        This function actually operates on one pixel at a time and is vectorized by numpy.

        Inputs
        ------
        nframes: float
            Number of unsaturated groups

        pix_cr_rate: float
            Rate of cosmic ray events per second per pixel (using size in microns)

        Raises:
            EngineInputError: _description_ EngineInputError: _description_

        Returns
        -------
        unsat_nframes: np.ndarray 
            2D array of unsaturated frames remaining after cosmic ray hits.
        """
        # this is the average fraction of the ramp that is lost upon a CR event
        if nframes < self.mingroups:
            # if for some reason (e.g. using single read noise model) ngroup is less than the minimum needed
            # to get a good ramp fit, then the whole ramp is lost upon a CR event
            ramp_frac = 1.0
        else:
            # the more groups you have per ramp, the less of the ramp you lose to CRs on average.
            # this formalism takes into account that there's always a fraction that's totally lost.
            # in the limit of infinite reads this converges on half the ramp since CRs are evenly
            # distributed.
            ramp_frac = 1.0 - 0.5 * (nframes - self.mingroups) / (nframes - 1.0)

        # the effective ramp exposure/measurement time, t_eff, is t_tot * (1 - ramp_frac * pix_cr_rate * t_ramp). We use the saturation
        # time as an approximation of the time during which a ramp can get damaged by a CR. This was previously more complex to
        # handle the case where a ramp is both saturated and hit by a CR, but that required a recalculation of the exposure time for
        # the unsaturated groups. We now simplify this (the difference is minimal and would add some complex logic now that the time
        # formulae are detector type dependent). This is simple.
        cr_nframes = (1.0 - ramp_frac * pix_cr_rate * self.exposure_spec.saturation_time) * nframes

        return cr_nframes

class H2RG(Detector_MultiAccum):

    def _get_sample(self):
        """
        H2RG and SiAs detectors use different parameters for slope variance
        (self.slope_variance, self.slope_variance_ta) and readnoise 
        (self.rn_slope_variance) (Issue #2996). Previously, the code tested to 
        see if special sample parameters existed, but now they always do.

        This is the H2RG version.
        """
        m = self.exposure_spec.nframe # This does not include skipped frames!
        tframe = self.exposure_spec.tframe
        return m, tframe

class SiAs(Detector_MultiAccum):

    def _get_sample(self):
        """
        H2RG and SiAs detectors use different parameters for slope variance
        (self.slope_variance, self.slope_variance_ta) and readnoise 
        (self.rn_slope_variance) (Issue #2996). Previously, the code tested to 
        see if special sample parameters existed, but now they always do.

        This is the SiAs version.
        """
        m = self.exposure_spec.nsample_total
        tframe = self.exposure_spec.tsample
        return m, tframe

class H4RG(Detector_UnevenMultiAccum):
    """
    Detector properties for H4RG detectors (Roman WFI).
    """
    def set_excess(self, flag):
        # this function turns excess noise on and off. H4RG detectors do not use excess noise parameters. Warn
        # if a user tries to turn it on.
        if flag:
            warnings.warn(f"{type(self).__name__} detectors do not have excess noise parameters, they cannot be turned on.")

#------------ Single accum ------------------

class Detector_SingleAccum(Detector):

    def set_excess(self, flag):
        # this function turns excess noise on and off. SingleAccum detectors do not use H2RG-style excess
        # noise parameters. Warn if a user tries to turn them on.
        if flag:
            warnings.warn(f"{type(self).__name__} detectors do not have excess noise parameters, they cannot be turned on.")

    def postflash_as_rate(self):
        """
        Postflash will not always be defined (and is only valid for some CCDs)

        Postflash must be a positive integer number of electrons.

        It's not really a rate, but to add it to the target flux we must treat its contribution as one, which
        means dividing it by the single exposure time - the flash happens before every independent exposure.
        """
        if (("postflash" in self.input_detector) and not self.__dict__.get("has_postflash", False)):
            raise EngineInputError(f"Detector does not support postflash.")
        postflash = self.input_detector.get("postflash", 0)
        if (not isinstance(postflash, int)) or (postflash < 0):
            raise EngineInputError(f"Postflash value {postflash} invalid. Must be positive integer.")

        # now convert to rate.
        postflash_rate = postflash / self.exposure_spec.exposure_time
        return postflash_rate

    def get_before_sat(self, rate):
        """
        This method returns a map of the maximum time each pixel can be exposed before saturation.

        This is considered the equivalent of the multiaccum groups-before-saturation, and serves
        a similar purpose: to tell the user the limits of what they can configure the detector to 
        do before saturation.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel
        """
        saturation_fraction = self.get_saturation_fraction(rate)

        time_before_sat = self.exposure_spec.saturation_time/saturation_fraction

        return time_before_sat

    def get_saturation_fraction(self, slope):
        """
        This method returns a map of the fraction of saturation that each pixel reaches, given the current
        exposure setup.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        """
        if isinstance(slope, dict):
            f = self.exposure_spec.saturation_time / self.fullwell
            return f * np.ones(shape=slope['fp_pix'].shape)
        else:
            # postflash is a single additive term taken at the end of the exposure,
            # not a rate. Thus, we need to determine the time for the signal, background,
            # and dark current to fill the REST of the well.
            if hasattr(self, "has_postflash") and "postflash" in self.input_detector:
                slope = slope - self.postflash_as_rate()
                fullwell = self.fullwell-self.input_detector["postflash"]
            else:
                fullwell = self.fullwell
            slope_array = slope.clip(MIN_CLIP, np.max([MIN_CLIP, np.max(slope)]))
            return self.exposure_spec.saturation_time / (fullwell / slope_array)

    def _get_saturation_mask(self, rate):
        """
        Compute a numpy array indicating pixels with (full) saturation (2), and no saturation (0).

        Parameters
        ----------
        rate: None or 2D np.ndarray
            Detector plane rate image used to build saturation map from

        Returns
        -------
        mask: 2D np.ndarray
            Saturation mask image
        """
        saturation_mask = np.zeros(rate.shape)

        saturation_fraction = self.get_saturation_fraction(rate)

        saturation_mask[(saturation_fraction > 1)] = 2

        return saturation_mask

    def get_unsaturated(self, rate):

        saturation_fraction = self.get_saturation_fraction(rate)

        unsat = ma.masked_greater(saturation_fraction, 1)

        return unsat

    def slope_variance(self, rate, unsat_ngroups):
        """
        Calculate the variance of a specified SingleAccum slope.
        Inputs
        ------
        rate: ndarray
            The measured slope of the ramp (ie, the rate) per pixel
            This can be a one-dimensional wavelength spectrum (for a 1D ETC)
            or a three-dimensional cube ([wave,x,y] for a 3D ETC).
        Returns
        -------
        slope_var: ndarray
            Variance associated with the input slope.
        slope_rn_var: ndarray
            The associated variance of the readnoise only
        """

        # The noise calculation depends on the pixel rate BEFORE IPC convolution. fp_pix_variance takes that
        # pre-IPC rate and scales it by the quantum yield and Fano factor to get the per-pixel variance in the
        # electron rate.
        
        # TODO: Determine if further changes to this equation need to be made. (JETC-2121) The reverse
        # calculation equation, derived from this noise computation, does not include Fano noise, nor does the
        # PyETC version. If we use a signal that includes fano noise, the reverse calculation does not reverse
        # the SNR calculation by a factor of more than 20%. Without it, the calculation is discrepant by ~1%.
        # More to the point, the Fano noise trends negative at longer wavelengths, which is almost certainly
        # unphysical. PyETC does not include Fano noise even though it's supposed to apply to detector
        # effects. It may yet be necessary to alter the Fano factor (set to 0) and quantum yield calculations
        # so that we CAN use an fp_pix_variance property that is never run through IPC convolution.

        variance_per_pix = rate['fp_pix'] / self.exposure_spec.total_exposure_time

        # SingleAccum detectors are read once, but signal variance and dark current are time-dependent.
        slope_rn_var = self.rn_variance() * np.ones_like(variance_per_pix) / self.exposure_spec.total_exposure_time**2
        slope_var = variance_per_pix + slope_rn_var

        # Noise is only valid when the pixel isn't saturated - on MultiAccum, that means operating on just the
        # unsaturated groups. In SingleAccum it means masking out the saturated pixels - they have no noise.
        # The default fill value for masked arrays is a finite number, so convert to ndarrays, and fill with
        # NaNs to make sure missing values are interpreted as truly undefined downstream.

        mask = self._get_saturation_mask(rate['fp_pix'])

        # Commenting this out so that 1D and scalar values can still be calculated for
        # saturated HST observations. This will not have an effect on the saturation maps,
        # the display of the 2D saturation, or the saturation warnings. It will, however,
        # prevent 1D plots from showing NaN in saturated areas. Because this change is
        # only to SingleAccum noise calculations, it will not affect JWST or Roman.
        # slope_var[mask==2] = np.nan 
        # slope_rn_var[mask==2] = np.nan

        return np.asarray(slope_var, dtype=np.float32), np.asarray(slope_rn_var, dtype=np.float32)

    def rn_variance(self):
        """
        Returns the variance due to read noise only.

        Returns
        -------
        rn: float
           Variance associated with the read noise only.

        """
        return self.rn**2 * self.exposure_spec.nsplit

    def calc_cr_loss(self, unsat_map, pix_cr_rate):
        """
        Calculate the effective loss of exposure time due to cosmic rays. This uses the cosmic ray (CR) event rate
        that is contained within the telescope configuration to calculate the odds of a cosmic rays
        hitting a pixel. Pixels that are hit by cosmic rays are assumed to have ramps that are valid before the
        CR, but not after. The exposure specification is used to adjust the expectation value of the exposure
        time (i.e. ngroups) for a CR-truncated ramp.

        See discussion and links in:

        - https://confluence.stsci.edu/pages/viewpage.action?spaceKey=JWST&title=2014-11-10+Effect+of+CRs+on+SNR
        - Robberto (2010) Technical Report JWST-STScI-001928

        for more information on implementation and the numbers used.

        There are singleaccum versions of the function im the above technical report, but fundamentally cosmic ray
        loss in SingleAccum detectors cannot be treated as statistical loss of exposure time: you either lose the
        entire pixel or you don't. Correspondingly, either the user can interpolate over (or coadd, otherwise 
        remove) the cosmic rays, in which case the full exposure time applies; or they can't, in which case the 
        observation is useless to them. 

        Parameters
        ----------
        unsat_map: int
            Map of unsaturated pixels

        pix_cr_rate: float
            Number of cosmic ray events per pixel area per second

        Returns
        -------
        cr_unsat: float
            This is the map of unsaturated pixels, scaled by the mean loss of time due to cosmic ray events.
        """

        # # if for some reason (e.g. using single read noise model) ngroup is less than the minimum needed
        # # to get a good ramp fit, then the whole ramp is lost upon a CR event
        # ramp_frac = 1.0

        # # the effective ramp exposure/measurement time, t_eff, is t_tot * (1 - ramp_frac * pix_cr_rate * t_ramp). We use the saturation
        # # time as an approximation of the time during which a ramp can get damaged by a CR. This was previously more complex to
        # # handle the case where a ramp is both saturated and hit by a CR, but that required a recalculation of the exposure time for
        # # the unsaturated groups. We now simplify this (the difference is minimal and would add some complex logic now that the time
        # # formulae are detector type dependent). This is simple.
        # cr_unsat = (1.0 - ramp_frac * pix_cr_rate * self.exposure_spec.saturation_time) * unsat_map

        return unsat_map

    def compute_scattering_factor(self, wave):
        """
        Compute the echelle scattered light parameter.

        Parameters
        ----------
        wave: float, numpy.ndarray
            Wavelength value(s) for which the scattering is desired.

        Returns
        -------
        scatter_factor

        """
        wave_scatter, mult_scatter = self.echelle_scattering
        scattering_val = np.interp(wave, wave_scatter, mult_scatter)
        scatter_factor = (1. / scattering_val - 1.)

        return scatter_factor

    def _get_variance_fudge(self, wave):
        """
        Some instruments (e.g. MIRI) have measured noise properties that are not well explained
        or modeled by Pandeia's current noise model. This method returns the multiplicative fudge factor
        that is to be applied to the pixel rate variance to match IDT expectations. The fudge factor can be
        chromatic so needs to accept an array of wavelengths to interpolate over. By default this returns
        a 1's array the same length as wave scaled by the 'var_fudge' scalar value.

        For single accum detectors, we keep this as identically 1.0 for now.

        Parameters
        ----------
        wave: 1D numpy.ndarray
            Wavelengths to interpolate chromatic variance over
        """
        return np.ones(len(wave))

    def set_time(self, new_time):
        # See also ExposureSpec.get_times()

        # measurement time is the time from the first measurement to the last read of the
        # exposure. HST has no skipped or dropped frames; this is simply the time for one
        # frame.
        self.exposure_spec.measurement_time = new_time / self.exposure_spec.nsplit

        # exposure time is the total time of one exposure (including any skipped reads,
        # reset reads, and the like - HST doesn’t have them.)
        self.exposure_spec.exposure_time = new_time / self.exposure_spec.nsplit
        
        # saturation time is the time the detector spends collecting photons between
        # resets, including skipped and dropped reads. It is the time that saturation
        # calculations depend on.
        self.exposure_spec.saturation_time = new_time / self.exposure_spec.nsplit
        
        # total exposure time is the time for the entire observation, including all
        # exposures and all the resets and skipped frames.
        self.exposure_spec.total_exposure_time = new_time

    def scale_exposure_time(self, extracted, calc_input):
        """
        Scale the exposure time to match the SNR we want.

        To do this requires a FULL understanding of the SingleAccum noise calculation done in the 
        Strategy and Report classes, because we need to implement its reverse here.

        This is only valid for SingleAccum calculations, where the function is:
        SNR = SR / sqrt(SR/T + BR/T + DR/T + ((Rn**2 + P) * nreads * npix)/T**2)
        and
        SR = Source Rate (electrons/s)
        BR = Background Rate (electrons/s)
        DR = Dark Current Rate (electrons/s)
        Rn = Readnoise (electrons)
        P = Postflash (electrons)
        nreads = Number of independent reads
        npix = Pixels over which Readnoise and Postflash are given (as they are per-pixel 
         values still)
        T = Time (s)
        (note the differences between this and the pyetc time formula - pyetc deals in counts; 
        Pandeia deals with rates)

        Per JETC-2065, solving for T (with an expected input SNR) yields this equation:
        0 = (Rn**2 + P)*npix*nreads + (SR + BR + DR) * T - (SR/SNR)**2 * T**2
        Which is a quadratic equation, solveable in the usual way.

        The complications to this (JETC-2121) are as follows:
        1. The Strategy outputs the extracted flux rate (SR) composed of just the source rate 
           itself, and a flux_plus_bg rate composed of the source, background, dark current, WFC3 
           thermal background, and postflash (SBDPR) - everything but readnoise, which only ever
           appears as a noise term. 
           - Postflash is included as a rate in SBDPR so that it appears in the 2D signal 
             products, but it is fundamentally not a rate and must be removed from SBDPR 
             (making SBDR) so it can be re-added correctly.
           - The other components (all of SBDR) are genuine rates.
           - The noise may include an ADDITIONAL term of STIS Echelle scattering added AFTER 
             Strategy.extract(), which will need to be accounted for. It is also a rate.
        2. The correlated noise extraction in the Strategy extracts the signal and noise through 
           (effectively) different apertures. If M is the mask array, the signal extraction is 
           M * SR and and the noise extraction is M * NR * M^T (see Strategy._error_sum), which 
           is equivalent to extracting the noise through M**2, which is smaller than M.
           Example:
           If this is the mask: 0    0.5  0
                                0.5  1    0.5
                                0    0.5  0       It covers 3 pixels (0.5 + 0.5 + 1 + 0.5 + 0.5)
           Then the noise aperture is the above matrix squared, or this:
                                0    0.25 0
                                0.25  1    0.25
                                0    0.25 0      It covers 2 pixels (0.25 + 0.25 + 1 + 0.25 + 0.25)
           This code computes the required noise aperture size (npix_noise) and mask correction 
           factors (mask_correction), which would be 0.66667 in my simple example here.
           - Signal is NOT spatially uniform (we cannot guess what the source flux present in those
             edge pixels is), so we really do need to use an actual extraction through the noise 
             aperture (extracted['extracted_source_noise']). We also have to remove the extracted 
             signal from the extracted flux_plus_bg_rate. (SBDR-SR = BDR)
           - Background, dark current, and thermal are all spatially uniform, and thus can be 
             handled by multiplying BDR by the mask_correction factor.
           - Postflash and Readnoise are also spatially uniform and given per pixel, and need to 
             be multiplied by the number of pixels in the noise aperture.
           - We do, however, still need the real extracted signal rate for the SR/SNR term.

        Parameters
        ----------
        extracted: dict
            Dictionary of output products produced by Strategy.extract
        calc_input: dict
            Dictionary of user inputs

        Returns
        -------
        time: float
            The exposure time to scale to.

        """
        wave_index = 0
        wave_pix = extracted['wavelength'] # at worst, this will be backwards from the actual wavelength grid
                                           # but we explicitly use min and max, so this should be fine.
        if len(wave_pix) > 1:
            wref = (wave_pix.max()+wave_pix.min())/2.
            user_wave = calc_input['strategy']['reference_wavelength']
            if user_wave is not None:
                if user_wave >= wave_pix.min() or user_wave <= wave_pix.max():
                    wref = user_wave
                # the Report version of this conditional produces a warning here (bad_waveref), but we won't 
                # bother now and let the Report produce it later.

            sbdpr = self.to_resels_scalar(wave_pix, extracted['extracted_flux_plus_bg_all'], wref) # (source + background + dark + postflash) rate
            sr = self.to_resels_scalar(wave_pix, extracted['extracted_flux'], wref) # source rate
            ns = self.to_resels_scalar(wave_pix, extracted['extracted_source_noise'], wref)
            if wref >= wave_pix[-2]:
                mask = extracted['masks_products'][-1]
            elif wref <= wave_pix[1]:
                mask = extracted['masks_products'][0]
            else:
                idx = bisect.bisect(wave_pix, wref)
                mask_0 = extracted['masks_products'][idx]
                mask_1 = extracted['masks_products'][idx+1]
                mask = mask_0 + (mask_1 - mask_0) * (wref - wave_pix[idx-1]) / (wave_pix[idx] - wave_pix[idx-1])
                
            # Echelle scattering, where it exists, must be accounted for in our noise formula
            if hasattr(self, "echelle_scattering"):
                scatter = self.compute_scattering_factor(wref)
            else:
                scatter = 0
        else:
            sbdpr = extracted['extracted_flux_plus_bg_all'][0] # (source + background + dark + postflash) rate
            sr = extracted['extracted_flux'][0] # source rate
            ns = extracted['extracted_source_noise'][0] # source rate through noise aperture
            mask = extracted['masks_products'][0]
            scatter = 0
            
        # The SNR we're shooting for, NOT the one we achieved
        snr = calc_input['configuration']['detector']['snr']

        # JETC-3020: Correction for HST spectroscopic calculations. The mask size we use
        # in the reverse calculation equation needs to be (pixels-per-resel) wide, because
        # the values it's using in the computation are in resels.
        pixels_per_resel = self.pixels_per_resel
        # this check excludes imaging modes (JETC-3020)
        if len(wave_pix) == 1:
            pixels_per_resel = 1

        # When background_subtraction is True, there will be negative pixels
        npix = np.sum(np.abs(mask)) * pixels_per_resel
        # the noise is computed over effectively fewer pixels, due to the fact that the
        # noise is squared and the edges of the mask are only partially transparent. But
        # that only applies in the cross-dispersion direction, so multiplying that value
        # by the width in unit pixels is appropriate.
        npix_noise = np.sum(mask**2) * pixels_per_resel

        mask_correction = npix_noise/npix
        # postflash is in the plus_background rate array but it is fundamentally not a
        # time-dependent term.
        postflash = self.input_detector.get("postflash", 0)
        sbdr = sbdpr - self.postflash_as_rate()*npix
        # we also need to remove the signal rate, because it is NOT a uniform term.
        bdr = (sbdr - sr) * mask_correction

        delta = (ns + bdr + (ns * scatter))**2 + 4 * (sr / snr)**2 * (self.rn**2 + postflash) * npix_noise * self.exposure_spec.nsplit

        time = ((ns + bdr + (ns * scatter)) + np.sqrt(delta)) / (2 * (sr / snr)**2) / pixels_per_resel

        return time
class CCD(Detector_SingleAccum):
    pass

class H1R(CCD):
    pass

class MAMA(Detector_SingleAccum):

    def __init__(self, config, webapp=False):
        """
        Sets the read noise correlation flag to False, instead of resorting to a
        definition in a config file (config files for MAMA detectors should not
        refer to a non-existent concept of read noise)
        """
        Detector_SingleAccum.__init__(self, config, webapp=False)

        self.rn_correlation = False

    def set_saturation(self, toggle):
        # MAMA detector has no concept of saturation, so this is irrelevant.
        pass

    def set_readnoise(self, flag):
        # this function turns readnoise on and off, but MAMAs do not have readnoise to turn on. Warn if a user
        # tries to turn it on.
        if flag:
            warnings.warn(f"{type(self).__name__} detectors do not have readnoise, it cannot be turned on.")

    def get_saturation_fraction(self, rate):
        """
        This method returns a map of the fraction of pixels that exceed
        the bright pixel rate threshold, given the current exposure setup.

        (we are reusing the 'saturation' word in the method's name. Maybe this
        should be revisited further down the road? For now, 'saturation' means,
        in the MAMA context, 'exceeds bright pixel threshold').

        Parameters
        ----------
        rate: ndarray
            The measured rate per pixel

        """
        arate = rate
        if isinstance(rate, dict):
            arate = rate['fp_pix']

        return np.zeros_like(arate)

    def rn_variance(self):
        """
        Returns the variance due to read noise only, which is identically zero
        for photon-counting sensors.

        Returns
        -------
        0.0: float
           The variance associated with the read noise on a MAMA detector is zero.

        """
        if self.rn != 0:
            warnings.warn(f"{type(self).__name__} detectors do not have readnoise, but it was set to {self.rn} e-/s. Setting to zero...")
            self.rn = 0
        return 0.0

class XDL(MAMA):
    pass



""" SEE NOTES IN pyetc repo  """

import pprint as pp
from xml.sax import saxutils

from pandeia.engine.helpers.bit.pyetc_util import log
from pandeia.engine.helpers.bit import input_conversion
from pandeia.engine.helpers.bit.perform_calculation import process_form_input

# ========================================================================

OUTPUT_MAP = {
    "BrightestPixelRate" : "brightest_pixel_rate",
    "ImageRate" : "detector_total_rate",
    "BackgroundRate" : "extracted_background_rate",

     'SegmentedCountRate.A': 'segment_a_total_rate',
     'SegmentedCountRate.B': 'segment_b_total_rate',
     'SegmentedCountRate.C': 'stripe_c_total_rate',

     # Forecasted mappings from Meg based on unimplemented pyetc functionality
     'SegmentedBrightPixel.A.value' : \
        lambda v, r: r["segmented_brightest_pixel_rate"]["a"][0],
     'SegmentedBrightPixel.B.value' : \
        lambda v, r: r["segmented_brightest_pixel_rate"]["b"][0],
}

def _has_segment_b(ei):
    """ Returns True if this is a COS calculation for which
    we would expect results for Segment B (not all do).
    Checks over the engine inputs (`ei`) dict passed in.
    """
    # sanity checks
    if ei['instrument'] != 'cos': raise ValueError('_has_segment_b called for non-COS calculation')
    for item in ('detector', 'grating', 'central_wavelength'):
        if item not in ei:
            raise ValueError(item+' not in engine inputs')

    segaltnames_key = (ei['grating'], ei['central_wavelength'])
    # check the COS instrument data config dict.  if segaltnames_key is in
    # it, then that will tell us if we should expect segment B.  if not, assume
    # it has a segment B
    detector = ei['detector']
    import pyetc.instruments.hst.cos as pihc # this is cached after 1st call
    cos_det_cfg = pihc.data.dconfig[detector]
    if segaltnames_key in cos_det_cfg['segaltnames']:
        segnames = cos_det_cfg['segaltnames'][segaltnames_key]
        return 'b' in segnames or 'B' in segnames
    else:
        return True

def get_and_map_requested_results(query, full_results, requests):
    """Map the variable names in `requests` from JETC nomenclature
    onto pyetc nomenclature and return them.  Return unavailable
    request variables as "undefined". The `query` var is passed in
    only to aid in debugging.

    dict(full_pyetc_results), list(requests) -->  dict(requested_jetc_results)
    """
    partial_results = {}
    for var in requests:
        mapped_var = None
        try:
            mapped_var = OUTPUT_MAP[var]
            # obsmode = full_results["engine_inputs"]["obsmode"]
            # grating = full_results["engine_inputs"].get("grating","none")
            if isinstance(mapped_var, str):
                partial_results[var] = full_results[mapped_var]
            else:
                partial_results[var] = mapped_var(var, full_results)
        except KeyError:
            partial_results[var] = "undefined"
            if mapped_var == None:
                log.error("Missing requested var from OUTPUT_MAP:", var)
            else:
                # The mapped_var is missing from full_results, so error, but if
                # this is segment B for COS, first see if we should even have it
                if full_results['engine_inputs']['instrument'] == 'cos' and \
                   var.startswith('Segmented') and '.B' in var and \
                   not _has_segment_b(full_results['engine_inputs']):
                    pass # this is OK - there is no segment B data
                else:
                    log.error("Missing requested mapped var from full results: "+str(mapped_var))
                    log.error("... query was:\n"+query+"\n... full_results is:\n"+str(full_results)+"\n")
    return partial_results

# ========================================================================

def get_partial_results(query, requests, full_results, message):
    """Return an XML <Response> associated with the result of a query.

    Pull out JETC named `requests` from pyetc:JETC mapped `full_results`.

    Include `query` in as the <query> element.

    Set <validRequest> to True IFF message is None.  Otherwise set
    <validRequest> to False and add <message>.

    Return a string.
    """
    partial_results = {}
    if message is not None:
        partial_results["message"] = message
        partial_results["validRequest"] = "false"
    else:
        requested = get_and_map_requested_results(query, full_results, requests)
        partial_results.update(requested)
        partial_results["validRequest"] = "true"
    partial_results["query"] = query
    xml = dict_to_xml("Response", partial_results)
    return str(xml) + "\n"

def dict_to_xml(name, dict_):
    """Serialize `dict_` as an XML element named `name`.

    >>> dict_to_xml("top_level", {
    ...        "this" : {
    ...            "nested":1,
    ...            "other":"str"
    ...        },
    ...        "that": 42
    ...    })
    '<top_level><this><other>str</other><nested>1</nested></this><that>42</that></top_level>'
    """
    return _make_element(name, _value_to_xml(dict_))

def _make_element(name, innards):
    """Wrap string `innards` in an XML begin/end tag."""
    return "<" + name + ">" + innards + "</" + name + ">"

def _value_to_xml(item):
    """Convert value `item` into XML,  recursively converting dictionaries
    into nested XML and XML-escaping the str() of non-dictionaries.
    """
    if isinstance(item, dict):
        xml = []
        for key, val in list(item.items()):
            xml.append(_make_element(key, _value_to_xml(val)))
        return "".join(xml)
    else:
        return saxutils.escape(str(item))

# ========================================================================

def compute_bright_object(web_inputs):
    """Given a set of pyetc web_inputs,  return pyetc .r dict with
    extra "engine_inputs" field.
    """
    engine_inputs = web_to_engine_inputs(web_inputs)

    results = enginecompute.compute(engine_inputs.copy(),
                                    safe_exceptions=False)
    results["engine_inputs"] = engine_inputs
    return results


def web_to_engine_inputs(web_inputs):
    """Convert complete pyetc web inputs into engine inputs.
    """
    form_dict = input_conversion.convert_dictionary(web_inputs)[0]
    instr = form_dict["instrument"]
    scimode = form_dict["science_mode"]

    engine_inputs = process_form_input(instr, scimode, form_dict, "./")[0]

    if log.VERBOSE_FLAG:
        log.verbose("\npyetc engine inputs:\n", pp.pformat(engine_inputs), "\n")

    return engine_inputs



import pandeia.engine.helpers.bit.instruments
#
# read a dictionary from one of the .dat files
def get_data( name ) :
    return pandeia.engine.helpers.bit.instruments.get_instrument_data( 'cos', name )

CONFIG = get_data("config")

def instrument_from_form(science_mode, form_input):
    """Function to extract values of interest from the web form
    data and return a smaller dictionary which will be used to
    update the engine_input dictionary.
    """
    if science_mode == 'spectroscopic' or science_mode == 'spectroscopicacq':
        obsmode = make_cos_spectroscopic_obsmode(form_input)
        disperser, detector, grating, central_wavelength, aperture = _get_cos_spec_parameters(form_input)
        idict = dict(disperser=disperser,
                     detector=detector,
                     grating=grating,
                     central_wavelength=central_wavelength,
                     aperture=aperture,
                     obsmode=obsmode)
        idict["wodict"] = wodict = {}
        # all cos detectors are mamas
        wodict["detector_type"] = 'mama'

        wodict["grating_info"] = {}
        wodict["grating_info"]["resolution"] = CONFIG[detector]["grating_resolution"][grating]
        
        if science_mode == 'spectroscopic':
            idict['obswave'] = form_input['obswave']
        else:
            #enforce a specacq prohibition
            if (form_input['disperser'] == 'nuv_0g230l' and
                form_input['nuvMode'] == 'ACQ/PEAKXD' and
                form_input['Stripe'] == 'C'):
                raise ValueError("You have selected grating G230L with Stripe C. This is not a valid mode for Spectroscopic Target Acquisition/PeakXD.")
            
    elif science_mode == 'imaging' or science_mode == 'targetacquisition':
        obsmode = make_cos_imaging_obsmode(form_input)
        mirror, aperture = _get_cos_imag_parameters(form_input)
        detector = 'nuv'
        idict = dict(obsmode=obsmode,
                     detector=detector,
                     mirror=mirror,
                     aperture=aperture)
        idict["wodict"] = wodict = {}
        # all cos detectors are mamas
        wodict["detector_type"] = 'mama'
        wodict["mirror"] = {"mirrora" : "Mirror A", "mirrorb" : "Mirror B"}[mirror]
        idict["filter"] = mirror
    else:
        raise ValueError('%s mode not implemented for cos'%science_mode)
    wodict["aperture"] = {
        "psa" : "Primary Science Aperture",
        "boa" : "Bright Object Aperture"
    }[aperture]
    return idict

def make_cos_spectroscopic_obsmode(idict):
    disperser, detector, grating,  central_wavelength, aperture = _get_cos_spec_parameters(idict)
    obsmode = ",".join(['cos', detector, grating, 'c'+str(central_wavelength), aperture])
    return obsmode

def make_cos_spectroscopicacq_obsmode(idict):
    return make_cos_spectroscopic_obsmode(idict)

def _get_cos_spec_parameters(idict):
    disperser = idict['disperser']
    detector, grating = disperser.split('_0',1)
    central_wavelength = idict[grating.upper() + '_CentralWavelength']
    aperture = idict['cosaperture0'].lower()
    return disperser, detector, grating, central_wavelength, aperture

def make_cos_imaging_obsmode(idict):
    mirror, aperture = _get_cos_imag_parameters(idict)
    obsmode = ",".join(['cos',mirror,'nuv',aperture])
    return obsmode

def make_cos_targetacquisition_obsmode(idict):
    return make_cos_imaging_obsmode(idict)

def _get_cos_imag_parameters(idict):
    mirror = idict['Mirror']
    aperture = idict['cosaperture0'].lower()
    return mirror, aperture



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .telescope import Telescope
from .instrument import Instrument
from .custom_exceptions import DataError
from .constants import DEFAULT_DISPERSION, PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS
import astropy.units as u
from astropy.table import Table
import synphot as syn
import stsynphot as st
import numpy as np
from numpy.polynomial import polynomial


class HST(Telescope):

    def get_ote_eff(self, wave):
        """
        Temporary function; allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)



class HSTInstrument(Instrument):

    """
    Generic HST Instrument class
    """
    def __init__(self, mode=None, config={}, **kwargs):
        telescope = HST()
        self.instrument_pars = {}
        # TODO HST has in general single accum detectors for which ngroup and nint do not make sense.
        self.instrument_pars['detector'] = ["nsplit"]
        self.instrument_pars['instrument'] = ["aperture", "disperser", "detector", "filter", "instrument", "mode"]
        self.api_parameters = list(self.instrument_pars.keys())

        # these are required for calculation, but ok to live with config file defaults
        self.api_ignore = ['dynamic_scene', 'max_scene_size', 'scene_size']
        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, **kwargs)
        # avoid doing the complex obsmode calculations multiple times
        self.obsmode = self.get_obsmode()
        self.rpower, self.dlds, self.dispwave = self.get_dispersion_products()

    def get_obsmode(self):
        if "imagingacq" in self.instrument['mode']:
            obsmode = self._obsmode_ta()
        elif "imaging" in self.instrument['mode']:
            obsmode = self._obsmode_imaging()
        elif "specacq" in self.instrument['mode']:
            obsmode = self._obsmode_specacq()
        elif "spec" in self.instrument['mode']:
            obsmode = self._obsmode_spec()
        elif "echelle" in self.instrument['mode']:
            obsmode = self._obsmode_spec()
        elif "ramp" in self.instrument['mode']:
            obsmode = self._obsmode_ramp()
            
        if "mjd" in self.instrument:
            mjd = self.instrument['mjd']
        else:
            mjd = self.telescope.mjd
        obsmode += ",mjd#{}".format(mjd)

        return obsmode

    def _obsmode_imaging(self):
        """
        Temporary function for basic imaging obsmodes, to be replaced as we work modes.
        These are likely not correct and not complete, but they should be functional.
        """
        return "{},{},{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['filter'])

    def _obsmode_spec(self):
        """
        Temporary function for basic spectroscopic obsmodes, to be replaced as we work modes.
        These are likely not correct and not complete, but they should be functional.
        """
        if "cenwave" in self.instrument:
            return "{},{},{},{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['disperser'], self.instrument['cenwave'])
        else:
            return "{},{},{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['disperser'])

    def _obsmode_specacq(self):
        """
        Base class defaults to spec mode
        """
        return self._obsmode_spec()

    def _obsmode_ta(self):
        """
        Base class defaults to imaging mode
        """
        return self._obsmode_imaging()

    def get_dispersion_products(self):
        """
        For HST we don't have separate files, we have the TRDS graph table
        and the default wavelength range being output
        """

        band = st.band(self.obsmode)
        dispwave = band.binset.to_value(u.micron)

        # code from pandeia_data/devtools/create_MIRI_LRS_dispersion.py
        dlds_int = np.gradient(dispwave)
        dlds_coeffs = polynomial.polyfit(dispwave,dlds_int,7)
        dlds = polynomial.polyval(dispwave, dlds_coeffs)

        r = dispwave/dlds
        return r, dlds, dispwave

    def get_dispersion(self, wave):
        """
        Return dispersion

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim dispersion data onto

        Returns
        -------
        dispersion: numpy.ndarray or float
            Dispersion as a function of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            dlds = self.dlds
            dlds = self._interp_refdata(wave, dlds, "dlds", default=DEFAULT_DISPERSION)
        else:
            dlds = 0.0
        return dlds

    def get_resolving_power(self, wave):
        """
        Return the resolving power

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate resolving power onto

        Returns
        -------
        rpower: numpy.ndarray or None
            Resolving power as a function of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            rpower = self.rpower
            rpower = self._interp_refdata(wave, rpower, "R", default=wave/(DEFAULT_DISPERSION*2))
        else:
            rpower = None
        return rpower

    def get_wave_pix(self):
        """
        Get the wavelength vector to convert pixel position to wavelength
        For HST, stsynphot and the graph table have already done this for us.

        Returns
        -------
        wavepix: numpy.ndarray
            Wavelength vector mapping pixels to wavelengths
        """
        return self.dispwave

    def get_detector_pixels(self, wave):
        """
        Read in detector pixel positions for each wavelength in wave_pix

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim pixel position data onto

        Returns
        -------
        pixels: numpy.ndarray or None
            Detector pixel positions corresponding to each element of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            pixels = np.arange(len(wave))
        else:
            pixels = None
        return pixels

    def get_wave_blaze(self):
        """
        Return the wavelength vector used in the grating efficiency (blaze) file

        Returns
        -------
        wave_blaze: numpy.ndarray
            Wavelength vector from the grating efficiency file
        """
        return self.dispwave


    def _interp_refdata(self, wave, data, name, default=None):
        """
        Read in requested reference file and return requested data interpolated onto wave

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array onto which the reference data is to be interpolated
        data: numpy.ndarray
            Key to get filename from self.paths
        name: string
            Name of the reference data; only used for the error message
        default: None or float
            Default value to return in case of missing file or column

        Returns
        -------
        ref: numpy.ndarray or float
            If file exists, return reference_data(wave). Else return default.
        """
        ref = None
        if data is not None:
            ref = np.interp(wave, self.dispwave, data)
        if ref is None and default is None:
            msg = "No reference data found and no default value supplied for %s." % (name)
            raise DataError(value=msg)
        elif ref is None and default is not None:
            ref = np.ones_like(wave)*default
        return ref

    def _get_cdbs(self, wave, key):
        """
        HST data is available in the $PYSYN_CBDS trees, so it should be loaded differently.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array onto which the throughput curve is to be interpolated
        key: str
            Key to get filename from the $PYSYN_CDBS tree

        Returns
        -------
        eff: numpy.ndarray or float
            If ref file exists, return efficiency(wave), else return 1.0
        """
        if 'None' not in key:
            bp = st.spectrum.band(key)
            eff = bp(wave*u.micron).value #don't want to pass the Quantity
        else:
            # If it's explicitly None, there is no element; pass through 1.0
            eff = np.ones_like(wave)*1.0
        return eff

    def get_filter_eff(self, wave):
        """
        Stub replacement to allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)

    def get_disperser_eff(self, wave):
        """
        Stub replacement to allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)

    def get_detector_qe(self, wave):
        """
        Stub replacement to allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)

    def get_internal_eff(self, wave, full_throughput=False):
        """
        This is the only temporary function that will actually do anything. It relies on
        stsynphot for graph table functionality. This does not run
        Instrument._get_throughput(), which would limit arrays to 32 bit. HST's
        high-resolution modes require 64-bit arrays.
        
        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto
        full_throughput: bool
            Dummy value that exists so the STIS-specific get_internal_eff can include slit
            losses.

        Returns
        -------
        eff: numpy.ndarray or float
            Disperser efficiency as a function of wave
        """
        bp = st.spectrum.band(self.obsmode)
        eff = bp(wave*u.micron).value #don't want to pass the Quantity

        return eff

    def get_quantum_yield(self,wave):
        """
        Compute detector quantum yield if the critical wavelength is defined. If not, 
        return unity (trivial yield).

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        quantum_yield: numpy.ndarray or float
            Quantum efficiency as a function of wave
        """
        if hasattr(self.the_detector, "critical_wavelength"):
            q_yield = self.the_detector.critical_wavelength / wave
        else:
            q_yield = 1
        fano_factor = (3.0 * q_yield - q_yield**2. - 2.) / q_yield

        return q_yield, fano_factor

class COS(HSTInstrument):

    """
    Special methods unique to HST COS
    """
    def __init__(self, mode=None, config={}, webapp=False, **kwargs):
        """
        Like JWST NIRSpec, COS needs to have the chip segments/stripes specially loaded
        """
        HSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)
        slit = self.instrument.get('slit', None)
        # Even COS imaging modes have slits, so a slit is not indicative of an issue.
        if self.projection_type != "image":
            try:
                slit_config = self.slit_config[slit]
            except KeyError as e:
                msg = "Configuration for slit {} not specified".format(slit)
                raise DataError(value=msg)
            self.segments = self.read_config_param(slit_config, "range")
        self._loadpsfs()

    def get_wave_range(self):
        if self.projection_type != "image":
            # get the wavelength range from the segments configuration
            disperser = self.instrument['disperser']
            cenwave = self.instrument['cenwave']
            segments = self.segments[disperser][cenwave]
            wmin = None
            wmax = None
            for segment in segments:
                if wmin == None or segment["wmin"] < wmin:
                    wmin = segment["wmin"]
                if wmax == None or segment["wmax"] > wmax:
                    wmax = segment["wmax"]
        else:
            aperture = self.instrument['aperture']
            slit = self.instrument['slit']
            wmin = self.range[aperture][slit]["wmin"]
            wmax = self.range[aperture][slit]["wmax"]

        range_dict = {'wmin': wmin, 'wmax': wmax}
        return range_dict

    def _get_psf_key(self):
        """
        The COS PSFs differ greatly by focus position, and we thus need to 
        know lifetime position (currently only LP4), disperser, and cenwave.
        """
        psf_key = self.instrument['aperture']

        if self.instrument['mode'] in ["fuv_spec","fuv_specacq", "nuv_spec", "nuv_specacq"]:
            psf_key = "{}{}{}".format(self.instrument['aperture'], self.instrument['disperser'], self.instrument['cenwave'])
        elif self.instrument['mode'] in ["nuv_imaging","nuv_imagingacq"]:
            psf_key = self.instrument['aperture']+self.instrument['slit']

        return psf_key

    def _obsmode_imaging(self):
        items = ['cos',self.instrument['aperture'],self.instrument['detector'], self.instrument['slit']]
        obsmode = ",".join(items)
        return obsmode

    def _obsmode_spec(self):
        items = ['cos',self.instrument['detector'],self.instrument['disperser'],str(self.instrument['cenwave']),self.instrument['slit']]
        obsmode = ",".join(items)
        return obsmode

    # specacq mode returns the same obsmode as spec - rely on base class directly
    def get_segment_indicies(self, wave):
        """
        Generic code to get the boundaries of a chip/stripe/segment in pixel coordinates, from our segment
        definitions.

        Parameters
        ----------
        wave: np.ndarray
            The wavelength array of the projected grid

        Returns
        -------
        segment_indices: list
            A list of dictionaries defining the segments in pixel coordinates.
        """
        disperser = self.instrument['disperser']
        cenwave = self.instrument['cenwave']

        segments = self.segments[disperser][cenwave]
        segment_indicies = []

        # loop over each segment
        for i in range(len(segments)):

            segname = segments[i]['name']

            # compute segment limits in X direction
            wmin = segments[i]['wmin']
            wmax = segments[i]['wmax']

            wmin_ind = np.searchsorted(wave,wmin, side="right")
            wmax_ind = np.searchsorted(wave,wmax, side="left")

            segment_dict = {"segname": segname, "wmin_ind": wmin_ind, "wmax_ind": wmax_ind}
            segment_indicies.append(segment_dict)

        return segment_indicies

    def get_chip_dimensions(self, wave):
        """
        COS FUV detector is comprised of two chips that have their own dark current, and require their own
        health and safety outputs; the spectrum is produced as a single array with a gap.

        Thus, we have both the complete dimensions of the chip (in pixel space) and the portion of the
        computed spectrum on that chip to consider.

        The COS NUV detector, on the other hand, is actually a single chip, but the optics split the spectrum
        into three stripes that are imaged separately over three different locations on the chip. We use this
        method then to create 3 "virtual segments" that help code downstream in its task of spliting up the
        spectrum into its three separate stripes. This has to be done so extractions and reporting can be
        performed on a per-stripe basis.

        In all of these, 
        
        - "name" is the detector name, 
        - "size" is the size of the section that collects dark current 
        - "bounds" is a dictionary containing the slices that will split up the flux (JUST the signal)
        into the appropriate pieces.

        background treatment is complex - if there is no slit, background will be assumed to cover the entire
        chip ("size"). If there IS a slit, if the slit is taller than the 2D detector image the background
        will be scaled to match. If it's not taller than the 2D image, we just assume the background array
        contains the entirety of the background flux.


        """
        nx = self.the_detector.pixels_x
        ny = self.the_detector.pixels_y

        if self.projection_type != "image":
            segment_indicies = self.get_segment_indicies(wave)
            dims = []
            if hasattr(self.the_detector, 'segnames'):
                for segment in segment_indicies:
                    dims.append({"name": segment["segname"], "size": {"x": nx, "y": ny},
                        "bounds": {"x": slice(segment["wmin_ind"],segment["wmax_ind"]), "y": slice(-1)}})

            elif hasattr(self.the_detector, 'stripenames'):
                stripe_ysize = int(ny / 3)
                for segment in segment_indicies:
                    dims.append({"name": segment["segname"], "size": {"x": nx, "y": stripe_ysize},
                        "bounds": {"x": slice(segment["wmin_ind"],segment["wmax_ind"]), "y": slice(-1)}})

            return dims

        return super().get_chip_dimensions(wave)

    def create_gap_mask(self, wave):
        """
        Use the slit segment configurations and a wavelength vector, wave, to build a masked array that masks
        out the location of the gaps.  Wavelengths between and including both gap endpoints will be masked
        out.

        As a reminder, the slit segments API is, as of 2022.0509:
        slit_config
         slit
          "range" <- self.segments
           subarray
            disperser
             filter or cenwave
              segment or stripe definitions.

        This code will not generate a slit mask unless a.) there is a slit configuration; b.) the actual
        defined slit is not None, and c.) if that slit definition includes a range item.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to construct mask from

        Returns
        -------
        mask: numpy.ma 1D masked array
            1D array masked at the locations within the configured detector gap and 1.0 elsewhere
        """
        mask = 1.0
        if hasattr(self, "segments"):
            segment_indicies = self.get_segment_indicies(wave)
            
            mask = np.zeros_like(wave)
            # we actually UNmask the non-gap area
            for segment in segment_indicies:
                mask[segment["wmin_ind"]:segment["wmax_ind"]] = 1.0

        return mask


class WFC3(HSTInstrument):

    """
    Special methods unique to HST WFC3
    """
    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        HSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)
        self._loadpsfs()

    def _obsmode_imaging(self):
        obsmode = ",".join(["wfc3",self.instrument['detector'],self.instrument['filter']])
        return obsmode

    def _obsmode_spec(self):
        obsmode = ",".join(["wfc3",self.instrument['detector'],self.instrument['disperser']])
        return obsmode

    def add_thermal(self):
        """
        This WFC3-specific functionality adds thermal background if the detector is the WFC3 IR
        detector.

        Per PyETC (engine/instrument.py line 272) thermback produces the integrated thermal 
        background per pixel, and it is to be treated as an additional source of flux akin
        to dark current.

        Returns
        -------
        thermal: float
            Scalar thermal background contribution
        """
        # Thermal is only a relevant parameter for WFC3's IR detector
        if self.instrument['detector'] == "ir":
            bp = st.spectrum.band(self.obsmode)
            # this is built out of integrating a spectrum in PHOTLAM/(arcsec**2) and
            # multiplying by area and (arcsec/pixel)**2, which should yield units of
            # PHOTLAM/(pixel**2) * cm^2.
            # where PHOTLAM is photons/sec/cm^2/Angstrom, and integrating should have 
            # removed the angstrom dependency. The final units should be photons/(sec 
            # pixel^2), which is being called counts/(s pix). That implies counts
            # is a unit of photons/pixel. 
            # we have this conversion: f_lambda = 1.5091905 * (flux / self.wave)
            # where flux is in mJy/pix and f_lambda is in photons/cm^2/sec/micron
            # We convert that value to electrons/s/pix/micron with the 
            # collecting area * ote_eff. And we don't have direct access to the HST 
            # OTA. 
            thermal = bp.thermback().value # counts/(pix sec)
            # We need electrons/(pix sec); which means converting counts to electrons.
            # Current WFC3IR gain is 2.5 e-/ADU, per 
            # https://hst-docs.stsci.edu/wfc3dhb/chapter-7-wfc3-ir-sources-of-error/7-2-gain 
            # (Accessed 2022.0209)
            # But, PyETC doesn't apply any gain, and all my calculations come out nearly 
            # 2.5x too high if I do.
            #thermal = syn.units.convert_flux(effwave*u.micron, bp.thermback(), u.electron/(u.pixel * u.second), area=self.telescope.coll_area)
        else:
            thermal = 0

        return thermal

class ACS(HSTInstrument):

    """
    Special methods unique to HST ACS
    """
    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        HSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)
        self._loadpsfs()

    def _obsmode_imaging(self):
        """
        Polarimeter options are already comma-separated and within the filter keyword.
        """
        obsmode = "{},{}".format(self.instrument['instrument'], self.instrument['detector'])
        if self.instrument['filter'] != "clear":
            obsmode += ",{}".format(self.instrument['filter'])
        return obsmode

    def _obsmode_spec(self):
        """
        Polarimeter options are already comma-separated and within the disperser keyword.
        """
        obsmode = "{},{},{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['disperser'])
        return obsmode
    
    def _obsmode_ramp(self):
        """
        Polarimeter options are already comma-separated and within the disperser keyword.
        """
        obsmode = "{},{},{}#{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['filter'], int(self.instrument['wavelength']*10000))
        return obsmode


class STIS(HSTInstrument):

    """
    Special methods unique to HST STIS
    """
    def __init__(self, mode=None, config={}, **kwargs):
        """
        STIS has additional configuration parameters, like fuv_glow_region and dark_level.
        Do the standard HSTInstrument setup, then read in the extra parameters.
        """
        super().__init__(mode, config, **kwargs)

        if 'fuvmama' in self.instrument['detector']:
            fuv_glow_key = 'dark_glow_region_{}'.format(config['detector']['fuv_glow_region'])
            fuv_glow_rate = self.detector_config['fuvmama'][fuv_glow_key]
            self.the_detector.dark_current = self.the_detector.dark_current + fuv_glow_rate
        self._loadpsfs()

    @property
    def projection_type(self):
        """
        Determine the appropriate projection type based on the configured instrument mode

        STIS is special, as its spec modes mix slit (spec) and slitless projections in the same
        mode. The selection is controlled by a "slitless_slits" list in the config.json file

        Returns
        -------
        proj_type: str
            The projection type, currently one of 'slitless', 'slitless_scan', 'spec', 'image', or 'image_scan'.
        """

        proj_type = super().projection_type
        if proj_type in ("spec", "slitless"):
            if self.instrument['slit'] in self.slitless_slits:
                proj_type = "slitless"
        return proj_type

    def _obsmode_imaging(self):
        items = ['stis',self.instrument['detector'], 'mirror']
        if self.instrument['filter'] is not None:
            items.append(self.instrument['filter'])
        obsmode = ",".join(items)
        return obsmode
    
    def _obsmode_ta(self):
        items = ['stis',self.instrument['detector'], 'mirror']
        if self.instrument['slit'] is not None:
            items.append(self.instrument['slit'])
        obsmode = ",".join(items)
        return obsmode

    def _obsmode_spec(self):
        """
        Much of the PyETC version of this code has to deal with the fact that the web interface
        does not contain the prefix character (i or c). In pandeia, the names of the cenwaves include
        the prefix.
        """
        disperser = self.instrument['disperser']
        detector = self.instrument['detector']
        obsmode = ",".join(['stis', detector, disperser])

        cenwave = self.instrument.get('cenwave',None)
        if cenwave != None:
            obsmode += ",{}".format(cenwave)
        else: #use default central wavelength if there is one
            try:
                cenwave = self.config_constraints['disperser'][disperser]['cenwaves'][0]
                obsmode += ",{}".format(cenwave)
            except (KeyError, IndexError):
                pass
                #otherwise leave it off
        
        slit = self.instrument.get('slit', None)
        if slit is not None:
            obsmode += ",{}".format(slit)

        return obsmode

    def get_internal_eff(self, wave, full_throughput=False):
        """
        Special STIS version of the get_internal_eff function. When a slit has both a
        filter AND a narrow slit, we need to remove the effect of the narrow slit so we
        don't double count it.
        
        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto
        full_throughput: bool
            Convenience for STIS Throughput plots, which DO want the slit losses.

        Returns
        -------
        eff: numpy.ndarray or float
            Disperser efficiency as a function of wave
        """
        slit = self.obsmode.split(",")[-2]

        if full_throughput:
            # For throughput plots, where we DO want to see the slit losses.
            bp = st.spectrum.band(self.obsmode)
            eff = bp(wave*u.micron).value # don't want to pass the Quantity
        elif slit in self.slitless_slits:
            # if the slit is slitless (only a filter) we don't need to worry about
            # double-counting light losses and can just use it.
            bp = st.spectrum.band(self.obsmode)
            eff = bp(wave*u.micron).value # don't want to pass the Quantity
        elif slit in self.unfiltered_slits:
            # if this is an unfiltered slit, we have already taken care of the slit losses
            # geometrically in ConvolvedSceneCube and should remove the slit parameter to
            # avoid double-counting.
            split_obsmode = self.obsmode.split(",")
            del split_obsmode[-2]
            obsmode = ",".join(split_obsmode)
            bp = st.spectrum.band(obsmode)
            eff = bp(wave*u.micron).value # don't want to pass the Quantity
        elif slit in self.filtered_slits:
            # if this is a filtered slit, we need to divide out the specific slit loss
            # component because we already handle that geometrically. 
            bp = st.spectrum.band(self.obsmode)

            # find the comp table
            comptable = st.getref()["comptable"]
            tab = st.stio.read_comptable(comptable)
            # find the STIS 52x0.05 slit file in the comp table
            slitfile = tab[[x.strip() == "stis_52x005" for x in tab["COMPNAME"]]]["FILENAME"][0]
            # load the slit bandpass
            slit_bp = syn.spectrum.SpectralElement.from_file(st.stio.irafconvert(slitfile))

            eff = bp(wave*u.micron).value
            slit_eff = slit_bp(wave*u.micron).value
            eff /= slit_eff

        return eff

    def apply_scattering(self, extracted_list):
        """
        Compute and apply the echelle scattered light parameter Will not affect any configuration that does
        not have defined echelle scattering 

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto
        rate: extracted target rate

        """
        if hasattr(self.the_detector, "echelle_scattering"):
            for idx in range(len(extracted_list)):

                scatter_rate = self.the_detector.compute_scattering_factor(extracted_list[idx]["wavelength"]) * extracted_list[idx]["extracted_source_noise"]

                # at this point, the noise is sqrt(variance), but we need to add this as a new poisson noise
                # term.
                extracted_list[idx]["extracted_noise"] = np.sqrt(extracted_list[idx]["extracted_noise"]**2 + scatter_rate / self.the_detector.exposure_spec.total_exposure_time)

        return extracted_list

    def get_global_scattering(self):
        """
        Return the global scattering rate, scattered portion as defined in the data
        """
        cenwave = self.instrument.get('cenwave', None)

        # No cenwave or cenwave = all will give no detector scattering, but PyETC doesn't make that option
        # possible.
        scatter_rate = 0
        if hasattr(self.the_detector, "echelle_global"):
            if cenwave in self.the_detector.echelle_global:
                scatter_rate = self.the_detector.echelle_global[cenwave] - 1

        return scatter_rate



#
# automatically generated by tools/form_field_table/gen_form_fields.py
#
field_list = [
    'AirglowStandard',
    'EarthshineMag',
    'EarthshineMult',
    'EarthshineSpec',
    'EarthshineStandard',
    'Mirror',
    'SNR',
    'Time',
    'ZodiDate',
    'ZodiDay',
    'ZodiDec',
    'ZodiMag',
    'ZodiMonth',
    'ZodiMult',
    'ZodiRA',
    'ZodiSpec',
    'ZodiStandard',
    'ZodiSun',
    'ZodiSunAttribute',
    'cosaperture0',
    'extractionRegionCircle',
    'extractionRegionExtendedCircle',
    'extractionRegionExtendedUserCircle',
    'extractionRegionExtendedUserCircle_min',
    'extractionRegionExtendedUserCircle_max',
    'extractionRegionUserCircle',
    'extractionRegionUserCircle_min',
    'extractionRegionUserCircle_max',
    'extractionRegionVariableCircle',
    'extractionRegionSquare',
    'extractionRegionExtendedSquare',
    'fband',
    'fbandpass',
    'fIndex',
    'fIsLambda',
    'fL1_center',
    'fL1_flux',
    'fL1_fwhm',
    'fL2_center',
    'fL2_flux',
    'fL2_fwhm',
    'fL3_center',
    'fL3_flux',
    'fL3_fwhm',
    'fOtherFile',
    'fRedshift',
    'fSpectrumCK',
    'fSpectrumPickles',
    'fStellar',
    'fUploadFile',
    'fbbtemp',
    'fbpgsfile',
    'fcalfile',
    'fphoenixfile',
    'fdiameter',
    'febmvtype',
    'febv',
    'fextinctiontype',
    'rn_flux_lambda',
    'rn_flux_lambda_units',
    'rn_flux_bandpass',
    'rn_flux_bandpass_units',
    'rn_lambda',
    'rn_lambda_units',
    'fftype_filters',
    'filter.ubvri',
    'filter.sloan',
    'filter.acs',
    'filter.nicmos',
    'filter.wfc3UVIS',
    'filter.wfc3IR',
    'fftype',
    'fnonstellar',
    'fsorigin',
    'fsourceType',
    'instrument',
    'science_mode',
    'simmode',
    'xRegionExtendedType',
    'xRegionType',
]



import pandeia.engine.helpers.bit.instruments

DETECTOR_TYPES = {'ccd' : 'ccd', 'fuvmama': 'mama', 'nuvmama': 'mama'}

#
# read a dictionary from one of the .dat files
def get_data( name ) :
    return pandeia.engine.helpers.bit.instruments.get_instrument_data( 'stis', name )

GRISMS = get_data("config")["gratings"]
CI_TABLE = get_data("config")["ci_table"]


def instrument_from_form(science_mode, form_input):
    """Function to extract values of interest from the web form
    data and return a smaller dictionary which will be used to
    update the engine_input dictionary.
    """
    # This gets rid of certain complaints from the migration code.
    binx    = form_input.get("binx",1) #get_default(form_input, "binx")
    biny    = form_input.get("biny",1)
    crsplit = form_input.get("crsplit",1)
    gain    = form_input.get("gain",1)
    #TODO: refactor so some common values can be set upfront
    if science_mode == 'spectroscopic':
        obsmode = make_stis_spectroscopic_obsmode(form_input)
        disperser, detector, grating, central_wavelength, aperture = _get_stis_spec_parameters(form_input)
        idict = dict(disperser=disperser,
                     detector=detector,
                     grating=grating,
                     aperture=aperture,
                     obsmode=obsmode,
                     crsplit=crsplit,
                     gain=gain,
                     binx=binx,
                     biny=biny,
                     )
        idict["wodict"] = wodict = {}

        idict['obswave'] = form_input['obswave']
        if central_wavelength != None:
            idict['central_wavelength'] = central_wavelength

        wodict["grating_info"] = GRISMS[grating.lower()]
        wodict["aperture"] = obsmode.split(',')[-1]
        wodict["detector_type"] =  DETECTOR_TYPES[detector].lower()

    elif science_mode == 'imaging' :

        # THESE MODES ARE STILL UNDER DEVELOPMENT.

        detector, aperture = _get_stis_imag_parameters(form_input)
        obsmode = make_stis_imaging_obsmode(detector, aperture)

        idict = dict(obsmode=obsmode,
                     detector=detector,
                     aperture=aperture,
                     crsplit=crsplit,
                     gain=gain,
                     binx=binx)

        idict["wodict"] = wodict = {}
        wodict["detector_type"] =  DETECTOR_TYPES[detector].lower()
        #idict['wodict'] = dict(detector_type=DETECTOR_TYPES[detector].lower())
        wodict["aperture"] = obsmode.split(',')[-1]
        idict["filter"] = wodict["aperture"].lower()
        wodict["obsmode"] = obsmode

    elif science_mode == 'targetacquisition':

        detector, aperture = _get_stis_imagacq_parameters(form_input)
        #the default extraction region has to be set to 32x32 for a PEAK ACQ aperture
        #this is currently being done by a hack in the extraction.py code which needs to
        #be changed at some point.

        aperture=form_input[aperture]
        ccdmode=form_input['ccdMode'].lower()
        point_target=(form_input['fsourceType'] == 'point')

        #The obsmode for imaging TA must specify the science aperture.
        obsmode = make_stis_targetacquisition_obsmode(detector,aperture,ccdmode,point_target)

        idict = dict(obsmode=obsmode,
                     detector=detector,
                     aperture=aperture,
                     ccdmode=ccdmode,
                     crsplit=crsplit,
                     gain=gain,
                     binx=binx)

        idict["wodict"] = wodict = {}
        wodict["detector_type"] = DETECTOR_TYPES[detector].lower()
        wodict["aperture"] = obsmode.split(',')[-1]
        idict["filter"] = wodict["aperture"].lower()
        wodict["obsmode"] = obsmode
        wodict["ccdmode"] = ccdmode

    else:
        raise ValueError('%s mode not implemented for stis'%science_mode)

    # this applies only to imaging and spectroscopic modes.
    idict["fuvglowregion"] = form_input.get("fuvglowregion", None)
    idict["ccddarklevel"] = form_input.get("ccddarklevel", None)

    return idict


def get_default(fdict, name):
    # Gets a default of 1 if name is not in the dictionary.
    try:
        return fdict[name]
    except KeyError:
        return 1


def make_stis_spectroscopic_obsmode(idict):
    disperser, detector, grating, central_wavelength, aperture = _get_stis_spec_parameters(idict)
            
    if central_wavelength != None:
        ci = CI_TABLE[grating][central_wavelength]
        obsmode = ",".join(['stis', detector, grating, ci + str(central_wavelength), aperture])

    else: #use default central wavelength if there is one
        try:
            grating_info = CI_TABLE[grating]
            central_wavelength = list(grating_info.keys())[0] #assume only 1 central wavelength in list
            designation=grating_info[central_wavelength]
            obsmode = ",".join(['stis', detector, grating, designation + str(central_wavelength), aperture])
        except KeyError:
            #otherise leave it off
            obsmode = ",".join(['stis', detector, grating, aperture])
    return obsmode


APERTURE_KEY = {
    "g750l" : "ccdaperture0",
    "g750m" : "ccdaperture0",
    "g430l" : "ccdaperture0",
    "g430m" : "ccdaperture0",
    "g705l" : "ccdaperture0",
    "g230lb" : "ccdaperture0",
    "g230mb" : "ccdaperture0",
    "g140l" : "fuvmamaaperture0",
    "g140m" : "fuvmamaaperture0",
    "g230l" : "nuvmamaaperture0",
    "g230m" : "nuvmamaaperture0",
    "e230m" : "e230maperture0",
    "e230h" : "e230haperture0",
    "e140m" : "e140maperture0",
    "e140h" : "e140haperture0",
    "prism" : "prismaperture0",
    }

def _get_stis_spec_parameters(idict):
    disperser = idict['disperser']
    detector, grating = disperser.split('_0',1)
    aperture = idict[APERTURE_KEY[grating]]
    try:
        central_wavelength = idict[grating.upper() + '_CentralWavelength']
    except KeyError:
        central_wavelength = None
    return disperser, detector, grating, central_wavelength, aperture

def make_stis_imaging_obsmode(detector, aperture=None):
    items = ['stis',detector,'mirror']
    if aperture is not None:
        items.append(aperture)
    obsmode = ",".join(items)
    return obsmode

def make_stis_targetacquisition_obsmode(detector,aperture,ccdmode,point_target):

    if "x" in aperture.lower() and "peak" in ccdmode.lower():
        #we need to put an s and remove the decimals
        dloc=aperture.split(".")
        aperture='s'+''.join(dloc)
    
    if "peak" in ccdmode.lower():
        if point_target:
            items = ['stis',detector,aperture]
        else:
            if "nd" in aperture.lower():
                items = ['stis',detector,aperture]
            else:
                items = ['stis',detector]        
    else:
        items = ['stis',detector,aperture]
        
    obsmode = ",".join(items)

    return obsmode

def _get_stis_imag_parameters(idict):
    detector = idict['detector']
    aperture = idict[detector + 'aperture0']
    return detector, aperture

def _get_stis_imagacq_parameters(idict):
    if idict["ccdMode"] == "ACQ":
        aperture = "ccdapertureACQ"
    else:
        aperture = "ccdapertureACQPEAK"
    detector = idict["detector"] = "ccd"
    return detector, aperture



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
from warnings import warn

from .custom_exceptions import DataError

class DetectorNoise(object):
    """
    This class provides the functionality to model the noise in a JWST observation

    Parameters
    ----------
    obs_signal: DetectorSignal or CombinedSignal instance
        The calculated signal at the detector plane
    observation: observation.Observation instance
        Configuration information for the Observation that generated ObsSignal
    validate: bool
        Keyword controlling whether intermediate products are saved for validation.
    """

    def __init__(self, obs_signal, observation, validate=False):
        self.warnings = {}
        self.validate = validate

        if self.validate:
            self.dist = obs_signal.dist
            self.grid = obs_signal.grid
            self.binning = obs_signal.binning
            # Are we dealing with an image or spectroscopic mode?
            self.projection_type = observation.instrument.projection_type
        self.det_mask = obs_signal.det_mask
        self.calculation_config = obs_signal.calculation_config


        self.the_detector = observation.instrument.the_detector

        # get saturation so we can calculate unsaturated groups
        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from detector_config
        if (self.calculation_config.effects['saturation'] if self.calculation_config.effects['saturation'] is not None else self.the_detector.saturation):
            self.the_detector.set_saturation(True)
        else:
            self.the_detector.set_saturation(False)

        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from detector_config
        if (self.calculation_config.noise['crs'] if self.calculation_config.noise['crs'] is not None else self.the_detector.crs):
            # if we are including CRs, we need a CR rate configured for the telescope. this rate is in events/s/cm^2/sr.
            if not hasattr(observation.instrument.telescope, 'cr_rate'):
                msg = "No cosmic ray rate defined for %s." % observation.instrument.telescope.tel_name.upper()
                raise DataError(value=msg)
            self.cr_rate = observation.instrument.telescope.cr_rate

            # we need physical pixel size to calculate incident area in sq cm
            if not hasattr(self.the_detector, 'pix_size_x') or not hasattr(self.the_detector, 'pix_size_y'):
                msg = "The physical pixel size of the detector in microns must be specified to calculate the CR rate."
                raise DataError(value=msg)
            self.pix_size_x = self.the_detector.pix_size_x
            self.pix_size_y = self.the_detector.pix_size_y

            # need to scale the rate by number of pixels affected by each event.
            if not hasattr(observation.instrument.telescope, 'cr_npixels'):
                msg = "No defined value for number of pixels affected per CR event for %s." % \
                      observation.instrument.telescope.tel_name.upper()
                raise DataError(value=msg)
            self.cr_npixels = observation.instrument.telescope.cr_npixels

            cr_rate = self.cr_rate * self.cr_npixels

            # cr_rate is in units of events/s/cm^2/sr.  convert pix_size in um to cm and
            # multiply by 4pi steradians to get cr_rate per pixel.
            self.pix_cr_rate = (self.pix_size_x * 1.0e-4 * self.pix_size_y * 1.0e-4) * 4.0 * np.pi * cr_rate
        else:
            self.pix_cr_rate = 0.0

        if self.validate:
            self.var_pix_list, self.stdev_pix_list, self.rn_var_pix_list, self.ff_pix_list = self.basic_source_noise(obs_signal)
        else:
            self.var_pix_list, self.stdev_pix_list, self.rn_var_pix_list = self.basic_source_noise(obs_signal)

        if self.validate:
            self.var_pix, self.stdev_pix, self.var_rn_pix, self.ff_pix = self.on_detector()
        else:
            self.var_pix, self.stdev_pix, self.var_rn_pix = self.on_detector()


    def on_detector(self, ff_pix=None):
        """
        Calculates the detector plane noise products

        Returns
        -------
        products: tuple
            detector_var - numpy.ndarray
                Full variance including all noise sources
            detector_stdev - numpy.ndarray
                Standard deviation (i.e. sqrt(detector_var))
            detector_rn_var - numpy.ndarray
                Variance stricly due to detector readnoise
            detector_ff_var - numpy ndarray
                Variance strictly due to flat field noise
        """
        aperture_sh = self.var_pix_list[0].shape
        n_apertures = len(self.var_pix_list)
        detector_shape = (aperture_sh[0] * n_apertures, aperture_sh[1])

        i = 0
        if self.validate:
            detected = [self.var_pix_list, self.stdev_pix_list, self.rn_var_pix_list, self.ff_pix_list]
        else:
            detected = [self.var_pix_list, self.stdev_pix_list, self.rn_var_pix_list]
        
        output = []
        for item in detected:
            i=0
            detector_item = np.zeros(detector_shape)
            for entry in item:
                detector_item[i * aperture_sh[0]:(i + 1) * aperture_sh[0]:] = entry

                i += 1
            output.append(self.det_mask * detector_item)

        return tuple(output)

    def basic_source_noise(self, obs_signal):
        """
        Calculate the noise using the full pixelated flux cube.

        Inputs
        ------
        obs_signal: DetectorSignal/CombinedSignal class instance
            Class containing the detector flux plane or plane cube

        Returns
        -------
        var_pix_list: List of ndarrays
            The detector variance plane or plane cube
        stdev_pix_list: List of ndarrays
            The detector standard deviation plane or plane cube (literally the square root of var_pix_list).
        rn_var_pix_list: List of ndarrays
            Variance strictly due to detector readnoise
        ffnoise_pix_list: List of ndarrays
            Variance strictly due to flat field noise
        """
        ff_electrons = self.the_detector.ff_electrons

        var_pix_list = []
        stdev_pix_list = []
        rn_var_pix_list = []
        ffnoise_pix_list = []

        for rate_plus_bg in obs_signal.rate_plus_bg_list:
            slope_var_pix, slope_rn_var_pix = self.get_slope_variance(rate_plus_bg)
            rate_per_pix = rate_plus_bg['fp_pix']
            """
            The flat field error is a division by ~1 (the flat field is normalized), with
            a variance of 1/ff_electrons. Note that the value of the flat field response
            is constant for multiple ramps and multiple integrations, so nramps > 1 does
            not decrease the residual flat field noise. Due to that, a user will either
            have to improve the flat field or dither with > 1 pixel offsets. The most
            apparent effect for everyday ETC use is that this sets an upper limit on the
            achievable signal-to-noise ratio.

            The pixel variance upon division with a normalized flat field constructed with
            ff_electrons (it is assumed that the flat field is ideal):

            s^2(R/FF) = s^2(R) + R^2/FF_electrons
            """
            var_pix = slope_var_pix / self.the_detector.exposure_spec.nramps
            rn_var_pix = slope_rn_var_pix / self.the_detector.exposure_spec.nramps

            # Add the flat field residual noise if requested. If there are multiple
            # exposures, decrease the noise to account for the effect of co-adding
            # exposures.
            # if (a if condition else b)
            # if the calculation_config value is True or False (edited by user), use True
            # or False. If it's none, use the instrument-team-defined default from
            # detector_config
            if (self.calculation_config.noise['ffnoise'] if self.calculation_config.noise['ffnoise'] is not None else self.the_detector.ffnoise):
                ffnoise_pix = (rate_per_pix ** 2 / ff_electrons) / self.the_detector.exposure_spec.exposures
                var_pix += ffnoise_pix
            else:
                ffnoise_pix = np.zeros_like(rate_per_pix)

            stdev_pix = np.sqrt(var_pix)

            var_pix_list.append(var_pix)
            stdev_pix_list.append(stdev_pix)
            rn_var_pix_list.append(rn_var_pix)
            ffnoise_pix_list.append(ffnoise_pix)

        if self.validate:
            products = var_pix_list, stdev_pix_list, rn_var_pix_list, ffnoise_pix_list
        else:
            products = var_pix_list, stdev_pix_list, rn_var_pix_list
        return products

    def get_slope_variance(self, rate):
        """
        Compute the slope variance, based on detector properties and
        exposure specifications.

        Inputs
        ------
        rate_per_pix: ndarray
           The measured slope of the ramp (ie, the rate) per pixel
           This can be a one-dimensional wavelength spectrum (for a 1D ETC)
        or a three-dimensional cube ([wave,x,y] for a 3D ETC).

        Returns
        -------
        slope_var: ndarray
           The associated variance of the measured slope
        slope_rn_var: ndarray
           The associated variance of the readnoise only
        """

        exp_pars = self.the_detector.exposure_spec


        unsat_ngroups = self.the_detector.get_unsaturated(rate['fp_pix_no_ipc'])
        # scale the unsaturated ngroups by the CR loss. While the fundamental thing being hit is individual unbinned pixels,
        # the math we've already done computes the effect per area (of unbinned pixel) and is thus still accurate
        #
        # The fact that unsat_ngroups is not an integer is intentional, and is the way the ETC statistically
        # accounts for CR losses.
        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from detector_config
        if (self.calculation_config.noise['crs'] if self.calculation_config.noise['crs'] is not None else self.the_detector.crs):
            unsat_ngroups = np.vectorize(self.the_detector.calc_cr_loss)(unsat_ngroups, self.pix_cr_rate)


        slope_var, slope_rn_var = self.the_detector.get_slope_variance(rate, unsat_ngroups)

        return slope_var, slope_rn_var

    def get_readnoise_slope_variance(self, rate):
        """
        Compute the variance of the read noise only, excluding all other noise contributions.

        Inputs
        ------
        none

        Returns
        -------
        var_rn: ndarray
            The variance of the readnoise given the current detector slope parameters. Note that
            if the readnoise is switched off, the associated variance will of course be zero.
        """

        unsat_ngroups = self.the_detector.get_unsaturated(rate['fp_pix_no_ipc'],
                                                        full_saturation=self.the_detector.mingroups)
        # scale the unsaturated ngroups by the CR loss.
        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from detector_config
        if (self.calculation_config.noise['crs'] if self.calculation_config.noise['crs'] is not None else self.the_detector.crs):
            unsat_ngroups = np.vectorize(self.the_detector.calc_cr_loss)(unsat_ngroups, self.pix_cr_rate)

        var_rn = self.the_detector.rn_variance(self.the_detector.rn, unsat_ngroups=unsat_ngroups)

        return var_rn
